/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "anchors.h"
#include "utils.h"
#include "generic_sum.h"
#include "anchor_sum.h"

void anchors_sum(double* X, double* Y, double* w, double* kparams, int NP, char* ker_type,
                       int NX, int NY, int D, double* mginfls, double epsilon_relative,
                       double epsilon_absolute, int influence_mode, int weights_mode) {

    int MIN_PARTICLES_PER_BOX = 1;
    Node *Xroot, *Yroot;
    tree_globals globals;
	globals.X = X;
	globals.Y = Y;
	globals.w = w;
	globals.D = D;
	globals.best = mginfls;
    globals.ibest = NULL;
    globals.influence_mode = influence_mode;
    globals.weights_mode = weights_mode;

    setup_kernel(ker_type);

    // initialize kernel
    kernel_init(kparams, NP);
    
    node_node_mindist = &anchors_node_node_mindist;
    node_node_maxdist = &anchors_node_node_maxdist;
    node_particle_mindist = &anchors_node_particle_mindist;
    node_particle_maxdist = &anchors_node_particle_maxdist;
    
    Xroot = build_anchors_hierarchy(X, D, NX, MIN_PARTICLES_PER_BOX);
    if (Y != X) {
        Yroot = build_anchors_hierarchy(Y, D, NY, MIN_PARTICLES_PER_BOX);
    } else {
        Yroot = Xroot;
    }
    generic_sum_decorate_x_tree(Xroot, w, &globals);
    generic_sum(&globals, Xroot, Yroot, epsilon_relative, epsilon_absolute);
    free_tree(Xroot, free_anchors);

    if (Y != X) {
        free_tree(Yroot, free_anchors);
    }

    (*kernel_destroy)();

}
