/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef ANCHORS_H_
#define ANCHORS_H_

#include "generic.h"

struct AnchorExtra {
    /* index (in X) of the anchor point. */
    int anchor;
    /* distances from the anchor to each point, sorted in decreasing order */
    double* distances;
    /* radius of the anchor (distance to furthest point) */
    double radius;
};
typedef struct AnchorExtra AnchorExtra;

// get the radius of the anchor
double anchor_radius(Node* a);
// build an anchors hierarchy for the data (X) in R^D, where there are N data points and the given
// leaf size and extra size for each anchor
Node* build_anchors_hierarchy(double* X, int D, int N, int MIN_PARTICLES_PER_BOX);
// free memory associated with the anchors hierarchy
void free_anchors(Node* node);

// returns the particle indexed by particle_index (in the anchor)
double* get_anchor_particle(double* X, Node* anchor, int particle_index, int D);
// returns the centre of the anchor
double* get_anchor_centre(double* X, Node* anchor, int D);

double anchors_node_node_mindist(Node* xnode, Node* ynode, tree_globals* globals);
double anchors_node_node_maxdist(Node* xnode, Node* ynode, tree_globals* globals);
double anchors_node_particle_mindist(Node* xnode, double* yparticle, tree_globals* globals);
double anchors_node_particle_maxdist(Node* xnode, double* yparticle, tree_globals* globals);

#endif /*ANCHORS_H_*/
