/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "boxtree.h"
#include "kdtree.h"
#include "utils.h"
#include "generic_max.h"
#include "boxtree_max.h"

void boxtree_max(double* X, double* Y, double* w, double* kparams, int NP, char* ker_type,
		              int NX, int NY, int D, int* mginds, double* mginfls,
		              int MIN_PARTICLES_PER_BOX, int BOXES_PER_LEVEL, int treetype, int influence_mode, int weights_mode) {
    Node *Xroot=NULL, *Yroot=NULL;
	tree_globals globals;
	globals.X = X;
	globals.Y = Y;
	globals.w = w;
	globals.D = D;
	globals.best = mginfls;
	globals.ibest = mginds;
	globals.influence_mode = influence_mode;
    globals.weights_mode = weights_mode;
  
    setup_kernel(ker_type);
  
    kernel_init(kparams, NP);
    
    node_node_mindist = &boxtree_node_node_mindist;
    node_node_maxdist = &boxtree_node_node_maxdist;
    node_particle_mindist = &boxtree_node_particle_mindist;
    node_particle_maxdist = &boxtree_node_particle_maxdist;

    if (treetype == BOXTREE) {
        Xroot = create_boxtree(X, NX, D, MIN_PARTICLES_PER_BOX, BOXES_PER_LEVEL);
    } else if (treetype == KDTREE) {
        Xroot = create_kdtree(X, NX, D, MIN_PARTICLES_PER_BOX);
	}

    if (Y != X) {
        if (treetype == BOXTREE) {
            Yroot = create_boxtree(Y, NY, D, MIN_PARTICLES_PER_BOX, BOXES_PER_LEVEL);
        } else if (treetype == KDTREE) {
            Yroot = create_kdtree(Y, NY, D, MIN_PARTICLES_PER_BOX);
        }
    } else {
        Yroot = Xroot;
    }

    generic_max_decorate_x_tree(Xroot, w, &globals);
    generic_max(&globals, Xroot, Yroot);

    free_tree(Xroot, free_boxtree);
    if (Y != X) {
        free_tree(Yroot, free_boxtree);
    }
    
    kernel_destroy();
}
