/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#ifndef BOXTREE_SUM_H
#define BOXTREE_SUM_H

/*
 * X = array of source particle data (size NX * D)
 * Y = array of target particle data (size NY * D)
 * w = array of weights (one per source particle)
 * kparams = kernel parameters (used to initialize kernel)
 * NP = number of kernel parameters
 * ker_type = type of kernel
 * NX = number of source particles
 * NY = number of target particles
 * D = dimension of particle space
 * mginfls = array of results (one per target particle) (I THINK)
 * BOXES_PER_LEVEL = arity??
 * epsilon_relative = desired relative error of approximation
 * epsilon_absolute = desired absolute error of approximation
 * treetype = type of tree
 */
void boxtree_sum(double* X, double* Y, double* w, 
		      double* kparams, int NP, char* ker_type,
                      int NX, int NY, int D, double* mginfls,
                      int arity, double epsilon_relative,
                      double epsilon_absolute, int treetype, int influence_mode, int weights_mode);

#endif /*BOXTREE_SUM_H*/
