/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "distance_transform.h"

INLINE double sqr(double a) {
  return a*a;
}

void printdbl(double *d, double *d2, int *idx, int n) {
  int i;
  for(i=0;i<n;i++)
    printf("%f [%f] org idx %d\n",d[idx[i]], (d2? d2[idx[i]] :0), idx[i]);
}

// dt of 1d function
void dt_max_irreg(double *X, double *Y, double *w, double* ker_pars, int NP, char* ker_type,
                  int NX, int NY, int D, int *mginds, double *mginfls) {
    double* logw;
    int i;
    double h;
	double *x_loc;
	double *para_bound;
	int *para_idx;
	int k, q;
	double *neglogw;
	int *oX;
	int *oY;


    if (strcmp(ker_type, "gaussian") != 0) {
        printf("kernel must be gaussian to do irreg DT!\n");
        return;
    }
    h = ker_pars[0];
  if(D != 1) {
    printf("D must be 1 to do irreg DT!\n");
    return;
  }
    logw = (double*) malloc(NX * sizeof(double));
    for (i = 0; i < NX; i++) {
        logw[i] = log(w[i]);
    }

  x_loc = (double*) malloc(NX * sizeof(double));
  para_bound = (double*) malloc((NX+1) * sizeof(double));
  para_idx = (int*) malloc(NX * sizeof(int));
  k = 0; q=0;

  // inverse sign of weights - we're doing a min here
  neglogw = (double*) malloc(NX * sizeof(double));

  // obtain sorted versions of X and Y
  oX = (int*) malloc(NX * sizeof(int));
  oY = (int*) malloc(NY * sizeof(int));

  for(q = 0; q < NX; q++)
    oX[q] = q;
  for(q = 0; q < NY; q++)
    oY[q] = q;

  compare_weights_vector = X;
  qsort(oX, NX, sizeof(int), compare_weights_asc);

  compare_weights_vector = Y;
  qsort(oY, NY, sizeof(int), compare_weights_asc);
  compare_weights_vector = NULL;

  /*
    printf("sorted X -------\n");
    printdbl(X, logw, oX, NX);
    printf("sorted Y -------\n");
    printdbl(Y, NULL, oY, NY);
  */

  /* initialize */
  for(q = 0; q < NX; q++)
    neglogw[q] = -logw[q];

  para_idx[0] = oX[0];
  x_loc[0] = X[oX[0]];
  para_bound[0] = -INF;
  para_bound[1] = +INF;

  for (q = 1; q < NX; q++) {
    double s  = ((h*h*neglogw[oX[q]]+sqr(X[oX[q]])) - 
		(h*h*neglogw[para_idx[k]]+sqr(x_loc[k]))) /
                    (2*(X[oX[q]]-x_loc[k]));
    while (s <= para_bound[k] && k > 0) {
      k--;
      //printf("%d\n",k);
      s  = ((h*h*neglogw[oX[q]]+sqr(X[oX[q]])) - 
		(h*h*neglogw[para_idx[k]]+sqr(x_loc[k]))) /
                    (2*(X[oX[q]]-x_loc[k]));

    }
    k++;
    para_idx[k] = oX[q];
    x_loc[k] = X[oX[q]];
    para_bound[k] = s;
    para_bound[k+1] = +INF;
  }

  /*
    printf("boundary of lower envelope:\n");
    for(q=0; q < NX && q < INF; q++)
    printf("bd %d at %f (corresp %d)\n", q, para_bound[q], para_idx[q]);
  */
  
  k = 0;
  for (q = 0; q < NY; q++) {
    while (para_bound[k+1] < Y[oY[q]]) {
      //printf("Xbd %f < yval %f\n", para_bound[k+1] , Y[oY[q]]);
      k++;
    }
//    mginfls[oY[q]] = exp(-1*sqr(Y[oY[q]]-x_loc[k])/(h*h) + logw[para_idx[k]]);
    mginfls[oY[q]] = w[para_idx[k]] * exp(-sqr(Y[oY[q]]-x_loc[k])/(h*h));
    if (mginds)
      mginds[oY[q]] = para_idx[k];
  }

  free(oX);
  free(oY);
  free(para_idx);
  free(para_bound);
  free(x_loc);
  free(neglogw);
}
