/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "anchor_max.h"
#include "boxtree_max.h"
#include "naive_max.h"
#include "anchor_sum.h"
#include "boxtree_sum.h"
#include "naive_sum.h"
#include "boxtree.h"
#include "distance_transform.h"
#include "gauss.h"

void nbody_sum(double* X, double* Y, double* w, double* kparams, int NP, char* ker_type, int NX,
         int NY, int D, double eps_rel, double eps_abs, char* algorithm, double* result, int influence_mode,
         int weights_mode) {
    if (eps_abs < 0) {
        eps_abs = 1e-100;
    }
    if (strcmp(algorithm, "anchors") == 0) {
        anchors_sum(X, Y, w, kparams, NP, ker_type, NX, NY, D, result, eps_rel, eps_abs, influence_mode, weights_mode);
    } else if (strcmp(algorithm, "kdtree") == 0) {
        boxtree_sum(X, Y, w, kparams, NP, ker_type, NX, NY, D, result, 2, eps_rel, eps_abs, KDTREE, influence_mode, weights_mode);
    } else if (strcmp(algorithm, "naive") == 0) {
        sum_naive(X, Y, w, kparams, NP, ker_type, NX, NY, D, result);
    } else if (strcmp(algorithm, "fgt") == 0) {
        gaussth(X, Y, w, kparams, NP, ker_type, NX, NY, D, result, eps_abs);
    }
}

void nbody_max(double* X, double* Y, double* w, double* kparams, int NP, char* ker_type,int NX,
         int NY, int D, int leafsize, char* algorithm, int* indices, double* results, int influence_mode, int weights_mode) {
    if (strcmp(algorithm, "anchors") == 0) {
        anchors_max(X, Y, w, kparams, NP, ker_type, NX, NY, D, indices, results, leafsize, influence_mode, weights_mode);
    } else if (strcmp(algorithm, "kdtree") == 0) {
        boxtree_max(X, Y, w, kparams, NP, ker_type, NX, NY, D, indices, results, leafsize, 2, KDTREE, influence_mode, weights_mode);
    } else if (strcmp(algorithm, "naive") == 0) {
        max_naive(X, Y, w, kparams, NP, ker_type, NX, NY, D, indices, results);
    } else if (strcmp(algorithm, "dt") == 0) {
        dt_max_irreg(X, Y, w, kparams, NP, ker_type, NX, NY, D, indices, results);
    }
}
