/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "dualtree.h"
#include <mex.h>
#include <matrix.h>

char* get_string_argument(const mxArray** args, int arg) {
  char* result;
  result = (char*) malloc(20 * sizeof(char));
  mxGetString(args[arg], result, 20); 
  return result;
}

void mexFunction(int nargout, mxArray *retval[], int nargin,  const mxArray *args[]) {
	const mxArray *mX, *mY, *mW, *mP;
    mxArray* mInfl;
    mxArray* mInds;
    double *X, *Y, *W, *P, *infl;
    double* dinds = NULL;
    int D, NX, NY, NP;
    char* ker;
    int influence_mode = 0, weights_mode = 0;
    char* ims;
    char* wms;
    char* type;
    char* structure;
    double eps_rel = 0;
    double eps_abs = 0;
	int* inds = NULL;
	int i;

    if (nargin < 9) {
        printf("Not enough arguments: need at least 9 (X,Y,w,P,kernel_name,influence_mode,weights_mode, \
                algorithm,computation_type[,relative_error,absolute_error]).\n");
        return;
    }

    mX = args[0];
    mY = args[1];
    mW = args[2];
    mP = args[3];

    X = mxGetPr(mX);
    Y = mxGetPr(mY);
    W = mxGetPr(mW);
    P = mxGetPr(mP);

    ker = get_string_argument(args, 4);
    ims = get_string_argument(args, 5);
    wms = get_string_argument(args, 6);
    structure = get_string_argument(args, 7);
    type = get_string_argument(args, 8);
    
    if (strcmp(ims, "positive") == 0) {
        influence_mode = POSITIVE_INFLUENCE;
    } else if (strcmp(ims, "negative") == 0) {
        influence_mode = NEGATIVE_INFLUENCE;
    } else if (strcmp(ims, "mixed") == 0) {
        influence_mode = MIXED_INFLUENCE;
    }
    
    if (strcmp(wms, "positive") == 0) {
        weights_mode = POSITIVE_WEIGHTS;
    } else if (strcmp(wms, "negative") == 0) {
        weights_mode = NEGATIVE_WEIGHTS;
    } else if (strcmp(wms, "mixed") == 0) {
        weights_mode = MIXED_WEIGHTS;
    }

    if (strcmp(type, "sum") == 0 && strcmp(structure, "naive") != 0) {
        eps_rel = *mxGetPr(args[9]);
        eps_abs = *mxGetPr(args[10]);
    }

    D = mxGetDimensions(mX)[0];
    NX = mxGetDimensions(mX)[1];
    NY = mxGetDimensions(mY)[1];
    NP = mxGetDimensions(mP)[1];
    
    mInfl = mxCreateDoubleMatrix(NY, 1, mxREAL);
    infl = mxGetPr(mInfl);
    retval[0] = mInfl;
    
	if (strcmp(type, "max") == 0) {
		mInds = mxCreateDoubleMatrix(NY, 1, mxREAL);
		retval[1] = mInds;
		dinds = mxGetPr(mInds);
		inds = (int*) malloc(NY * sizeof(int));
		for (i = 0; i < NY; i++) {
			inds[i] = -1;
		}
	}

    if (strcmp(type, "sum") == 0) {
        nbody_sum(X, Y, W, P, NP, ker, NX, NY, D, eps_rel, eps_abs, structure, infl, influence_mode, weights_mode);
    } else if (strcmp(type, "max") == 0) {
        nbody_max(X, Y, W, P, NP, ker, NX, NY, D, 25, structure, inds, infl, influence_mode, weights_mode);
    }

	if (strcmp(type, "max") == 0) {
		for (i = 0; i < NY; i++) {
			dinds[i] = (double) inds[i];
		}
		free(inds);
	}

	free(ker);
	free(type);
	free(structure);
    
}
