/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "generic.h"
#include "definitions.h"

INLINE void free_tree(Node* tree, void (*free_extra)(Node* node)) {
    int i;
    // free the children first...
    for (i = 0; i < tree->Nchildren; i++) {
        free_tree(tree->children[i], free_extra);
    }    
    if (tree->children) {
        free(tree->children);
    }
    if (tree->indices) {
        free(tree->indices);
    }
    // call the user's function to free any extra resources...
    if (free_extra) {
        free_extra(tree);
    }
    free(tree);
    tree = NULL;
}

INLINE int is_leaf(Node* node) {
    return !(node->children);
}

//INLINE int get_N(Node* node) {
//    return node->N;
//}

INLINE double get_weight(Node* xnode, int index, tree_globals* globals) {
    return globals->w[xnode->indices[index]];
}

//INLINE int N_children(Node* node) {
//    return node->Nchildren;
//}

INLINE Node* get_child(Node* node, int c) {
    return node->children[c];
}

INLINE int* get_indices(Node* node) {
    return node->indices;
}

INLINE int get_unsorted_index(Node* node, int indx) {
    return node->indices[indx];
}

INLINE void set_best_influence(Node* ynode, int index, int ibest, double best, tree_globals* globals) {
    globals->best[ynode->indices[index]] = best;
    // used for max 
    if (globals->ibest) {
        globals->ibest[ynode->indices[index]] = ibest;
    }
}

INLINE void set_maxweight(Node* node, double maxweight) {
    node->maxweight = maxweight;
}

INLINE void set_minweight(Node* node, double minweight) {
    node->minweight = minweight;
}

INLINE void set_sumposweight(Node* node, double sumposweight) {
    node->sumposweight = sumposweight;
}

INLINE void set_sumnegweight(Node* node, double sumnegweight) {
    node->sumnegweight = sumnegweight;
}

INLINE double get_maxweight(Node* node) {
    return node->maxweight;
}

INLINE double get_minweight(Node* node) {
    return node->minweight;
}

INLINE double get_sumposweight(Node* node) {
    return node->sumposweight;
}

INLINE double get_sumnegweight(Node* node) {
    return node->sumnegweight;
}

INLINE void set_sorted_indices(Node* node, int* sorted) {
    if (node->indices) {
        free(node->indices);
    }
    node->indices = sorted;
}

INLINE double* get_x_particle(Node* xnode, int index, tree_globals* globals) {
    return globals->X + xnode->indices[index] * globals->D;
}

INLINE double* get_y_particle(Node* ynode, int index, tree_globals* globals) {
    return globals->Y + ynode->indices[index] * globals->D;
}

//INLINE int all_particles_are_equal(double* X, int D, Node* node) {
//	double* p1 = X + node->indices[0] * D;
//	double* p;
//	int i;
//	int d;
//	for (i = 1; i < node->N; i++) {
//		p = X + node->indices[i] * D;
//		for (d = 0; d < D; d++) {
//			if (p1[d] != p[d]) {
//				return 0;
//			}
//		}
//	}
//	return 1;
//}
