/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "generic.h"
/**
   \file
   The fast dual-tree sum-product algorithm.

   \ingroup sum
*/

/**
   The fast dual-tree sum-product algorithm.

   \param Xroot root of the X (source particles) tree.
   \param Yroot root of the Y (target points) tree.
   \param epsilon_relative <b>CURRENTLY NOT SUPPORTED</b>.
   \param epsilon_absolute maximum allowed absolute error per estimate.
*/
void generic_sum(tree_globals* globals, Node* Xroot, Node* Yroot, double epsilon_relative,
		 double epsilon_absolute);

/**
   A convenience function for 'decorating' the X (source) tree with
   extra information used during the <tt>generic_sum</tt> routine.

   Call this function on the X tree before calling <tt>generic_sum</tt>.

   \param node the root of the X tree.
   \param w the array of weights
*/
void generic_sum_decorate_x_tree(Node* node, double* w, tree_globals* globals);


#if defined(NEG)
void generic_sum_decorate_x_tree_neg(Node* node, double* w, tree_globals* globals);
#endif

///**
//   An accounting variable for recording the amount of time (in ms) spent in the ALL phase
//   of the algorithm.
//*/   
//extern double time_phase_sum;
///**
//   An accounting variable for recording the amount of time (in ms) spent in the SUM phase
//   of the algorithm.
//*/   
//extern double time_phase_all;

