/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "kdtree.h"

/**
   \internal 
   Builds a KD-Tree recursively.  Called by \code create_kdtree .
*/
void kdtree_recursive(double* X, int N, int D, Node* node,
		      int leafsize) {
    int d, i;
    int bestd = -1;
    double maxrange;
	BoxExtra* be;

    // Base case: this is a leaf node.
    if (N <= leafsize) {
        return;
    }

    // Recursive case: find the dimension with largest range (max - min).
    maxrange = -1.0;
    be = (BoxExtra*) node->extra;
    for (d=0; d<D; d++) {
        double range = be->highvals[d] - be->lowvals[d];
        if (range > maxrange) {
            maxrange = range;
            bestd = d;
        }
    }
    // Create child nodes and place points in them.
    place_in_boxes(node, N, X, D, bestd, 2);
    // Recurse!
    for (i=0; i<node->Nchildren; i++) {
        kdtree_recursive(X, node->children[i]->N, D, node->children[i], leafsize);
    }

}

Node* create_kdtree(double* X, int N, int D, int leafsize) {
    Node* root;
    int i;

    // Here we create and initialize the root node.
    root = boxtree_create_node(N, D);
    for (i=0; i<N; i++) {
        root->indices[i] = i;
    }
    boxtree_tighten_box_limits(root, X, D);
    kdtree_recursive(X, N, D, root, leafsize);
    return root;
}
