/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
#ifndef KDTREE_H
#define KDTREE_H

#include "boxtree.h"

/**
 * Create a KD-Tree with:
 *   \param X the data points (see boxtree.h for information about how the data is laid out.)
 *   \param N the number of data points.
 *   \param D the dimensionality of the data points.
 *   \param leafsize the maximum number of points that can be contained in leaf nodes
 *         of the tree.
 *
 * \return pointer to the root of the tree.
 *
 * Note that "kdtree" is typedef'd to be a
 * BoxTree; see boxtree.h for the definition of this structure.
 */
Node* create_kdtree(double* X, int N, int D, int leafsize);

#endif /*KDTREE_H*/
