/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef KERNEL_H
#define KERNEL_H

/*
Provides the interface needed to interact with a custom kernel function
 - Implemented using function pointers
*/

/**
   Kernel Parameters
*/

int pnum;         // number of parameters
char* ker_type;   // the kernel type
double* ker_pars; // array of kernel parameters

/* Guaranteed to be called exactly once at the outset */
void (*kernel_init)(double* kernel_params, int n_par);

/* Guaranteed to be called exactly once at the outset */
void (*kernel_destroy)();

/**
   Given a weight and a distance,
   compute the influence of a particle of the given weight at
   the given distance.   
*/
double (*kernel_influence)(double weight, double dist);

double (*kernel_distance)(double* x, double* y, int N);

int (*kernel_increasing)();

#endif

