/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "kernel.h"
#include "kernel_laplacian.h"

// initialize the kernel parameters
// NP = number of parameters
void laplace_kernel_init(double* kernel_params, int NP) {
	double h;
    pnum = 1;
    ker_pars = (double*) malloc(pnum * sizeof(double));
    if (pnum != NP){
        printf("Wrong number of parameters: %d\n", NP);
        return;
    }
    h = kernel_params[0];
    ker_pars[0] = h;
}

void laplace_kernel_destroy() {
    // do nothing
    free(ker_pars);
    ker_pars = NULL;
}

double laplace_influence(double weight, double dist) {
    return weight * exp(-dist / (ker_pars[0]));
}

double laplace_distance(double* x, double* y, int D) {
    int i;
    double result = 0;
    for (i = 0; i < D; i++) {
      result += dabs(x[i] - y[i]);
//      result += square(x[i] - y[i]);
    }
    return result;
//    return sqrt(result);
}

int laplace_increasing() {
    return 0;
}
