/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "kernel.h"
#include "kernel_lpq.h"

// initialize the kernel parameters
// NP = number of parameters
void lpq_kernel_init(double* kernel_params, int NP) {
	double p, q, increasing;
    pnum = 3;
    ker_pars = (double*) malloc (pnum * sizeof(double));
    if (pnum != NP){
        printf("Wrong number of parameters: %d\n", NP);
        return;
    }
    p = kernel_params[0];
    q = kernel_params[1];
    increasing = kernel_params[2];
    printf("p = %f, q = %f, increasing = %f\n", p, q, increasing);
    ker_pars[0] = p;
    ker_pars[1] = q;
    ker_pars[2] = increasing;
    if (ker_pars[2]) {
        printf("increasing!\n");
    } else {
        printf("decreasing!\n");
    }
}

void lpq_kernel_destroy() {
    free(ker_pars);
    ker_pars = NULL;
}

double lpq_influence(double weight, double dist) {
    if (ker_pars[2]) {
        return weight * (1 + pow(dist, ker_pars[1]));
    } else {
        return weight / (1 + pow(dist, ker_pars[1]));
    }
}

double lpq_distance(double* x, double* y, int D) {
    int i;
    double result = 0;
    for (i = 0; i < D; i++) {
        result += pow(dabs(x[i] - y[i]), ker_pars[0]);
    }
    return pow(result, 1.0 / ker_pars[0]);
}

int lpq_increasing() {
    return (int) ker_pars[2];
}
