/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "utils.h"
#include "kernel.h"

/*
 * X = array of source particle data (size NX * D)
 * Y = array of target particle data (size NY * D)
 * w = array of weights (one per source particle)
 * h = bandwidth
 * NX = number of source particles
 * NY = number of target particles
 * D = dimension of particle space
 * ibests = array of results (one per target particle) giving the index of the source particle
 *          that exerts the maximum influence on the target particle
 * dbests = same as ibests but using a double instead of an int
 * bests = array of results (one per target particle) giving the value of the maximum influence
 *         exerted on the target particle by a single source particle
 */

void max_naive(double *X, double *Y, double *w, double* kparams, int NP,
    char* ker_type, int NX, int NY, int D, int* ibests, double* bests) {
    int x, y;
    
    setup_kernel(ker_type);

    kernel_init(kparams, NP);
    
    for (y=0; y<NY; y++) {
        double best = -INF;
        int ibest = -1;
        for (x=0; x<NX; x++) {
            double dist = kernel_distance(X + x*D, Y + y*D, D);
            double infl = kernel_influence(w[x], dist);
            if (infl > best) {
                best = infl;
                ibest = x;
            }
        }
        if(ibests) {
            ibests[y] = ibest;
        }
        bests[y] = best;
    }
    
    kernel_destroy();
}
