/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "naive_sum.h"
#include "kernel.h"
#include "utils.h"

/*
 * X = array of source particle data (size NX * D)
 * Y = array of target particle data (size NY * D)
 * w = array of weights (one per source particle)
 * kparams = kernel parameters (used to initialize kernel)
 * NP = number of kernel parameters
 * ker_type = type of kernel
 * NX = number of source particles
 * NY = number of target particles
 * D = dimension of particle space
 * bests = array of results (one per target particle)
 */
void sum_naive(double *X, double *Y, double *w, double* kparams, int NP,
    char* ker_type, int NX, int NY, int D, double* bests) {
    int x, y;
    
    setup_kernel(ker_type);

    kernel_init(kparams, NP);

    for (y=0; y<NY; y++) {
        double sum = 0.0;
        for (x=0; x<NX; x++) {
            double dist;
            double infl;
            dist = kernel_distance(&X[x*D],&Y[y*D],D);
            infl = kernel_influence(w[x], dist);
            sum += infl;
        }
        bests[y] = sum;
    }
    kernel_destroy();
}
