/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "nodelist.h"
#include "definitions.h"

INLINE nodelist* nodelist_new() {
    nodelist* nl = (nodelist*) malloc(sizeof(nodelist));
    nl->size = 0;
    nl->capacity = 100;
    nl->array = (Node**) malloc(nl->capacity * sizeof(Node*));
    return nl;
}

INLINE void nodelist_free(nodelist* nl) {
    free(nl->array);
    free(nl);
    nl = NULL;
}

INLINE void nodelist_remove_after_index(nodelist* nl, int start) {
    nl->size = start;
}

INLINE int nodelist_size(nodelist* nl) {
    return nl->size;
}

INLINE void clear_nodelist(nodelist* nl) {
    nl->size = 0;
}

INLINE void nodelist_increase_capacity(nodelist* nl) {
    nl->capacity *= 2;
    nl->array = (Node**) realloc(nl->array, nl->capacity * sizeof(Node*));
}

INLINE void nodelist_append_list(nodelist* dest, nodelist* src) {
    int i;
    if (src->size == 0) {
        return;
    }
    while (dest->size + src->size > dest->capacity) {
        nodelist_increase_capacity(dest);
    }
    for (i = 0; i < src->size; i++) {
        dest->array[dest->size + i] = src->array[i];
    }
    dest->size += src->size;
    clear_nodelist(src);
}

INLINE Node* nodelist_get(nodelist* nl, int index) {
    return nl->array[index];
}

INLINE void nodelist_remove(nodelist* nl, Node* an) {
    int i;
    int j;
    for (i = 0; i < nl->size; i++) {
        if (nl->array[i] == an) {
            for (j = i + 1; j < nl->size; j++) {
                nl->array[j - 1] = nl->array[j];
            }
            nl->size--;
            return;
        }
    }
}

INLINE void nodelist_copy(nodelist* nl, int num, Node** dest) {
    int i;
    for (i = 0; i < num; i++) {
        dest[i] = nl->array[i];
    }
}

INLINE void nodelist_append(nodelist* nl, Node* an) {
    if (nl->size == nl->capacity) {
        nodelist_increase_capacity(nl);
    }
    nl->array[nl->size] = an;
    nl->size++;
}
