/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "pqueue.h"
#include "definitions.h"

pqueue* pq_new(int max_order, int hopeless) {
    pqueue* pq = (pqueue*) malloc(sizeof(pqueue));
    // note: this might not be big enough for anchors
    int array_size = max_order - hopeless + 10;
	int i;
    pq->heads = (bound_node**) malloc(array_size * sizeof(bound_node*));
    for (i = 0; i < array_size; i++) {
        pq->heads[i] = NULL;
    }
    pq->num_orders = array_size;
    pq->size = 0;
    pq->hopeless = hopeless;
    pq->max_index = -1;
    return pq;
}

void pq_enqueue(pqueue* pq, bound_node* bna) {
    int index = bna->order - pq->hopeless;
    if (pq->heads[index] == NULL) {
        if (index > pq->max_index) {
            pq->max_index = index;
        }
    }
    bna->next = pq->heads[index];
    pq->heads[index] = bna;
    pq->size++;
}

bound_node* pq_pop(pqueue* pq) {
	bound_node* rtn;
    if (pq->size == 0) {
        return NULL;
    }
    rtn = pq->heads[pq->max_index];
    pq->heads[pq->max_index] = rtn->next;
    if (rtn->next == NULL) {
        if (pq->size == 1) {
            pq->max_index = -1;
        } else {
            int i = pq->max_index;
            while (pq->heads[i] == NULL) {
                i--;
            }
            pq->max_index = i;
        }
    }
    rtn->next = NULL;
    pq->size--;
    return rtn;
}

void pq_print(pqueue* pq) {
    int i;
    printf("size: %d\n", pq->size);
    for (i = pq->num_orders - 1; i >= 0; i--) {
        bound_node* bn = pq->heads[i];
        while (bn != NULL) {
            printf("BN: level %d, order %d, lower %f, upper %f\n", i, bn->order, bn->lower, bn->upper);
            bn = bn->next;
        }
    }
}

int pq_real_size(pqueue* pq) {
    int i;
    int size = 0;
    for (i = pq->num_orders - 1; i >= 0; i--) {
        bound_node* bn = pq->heads[i];
        while (bn != NULL) {
            size++;
            bn = bn->next;
        }
    }
    return size;
}

void pq_free(pqueue* pq) {
    int i;
    for (i = pq->num_orders - 1; i >= 0; i--) {
        bound_node* bn = pq->heads[i];
        while (bn != NULL) {
            bound_node* toFree = bn;
            bn = bn->next;
            free(toFree);
        }
    }
    free(pq->heads);
    free(pq);
    pq = NULL;
}
