/*
 * N-Body Methods
 * Copyright (C) 2006 Dustin Lang, Mike Klass, Firas Hamze, Anthony Lee
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef UTILS_H_
#define UTILS_H_

#include "definitions.h"
#include "generic.h"
#ifdef WINDOWS
#include <windows.h>
#endif
#ifdef LCC
struct timeval {
    long tv_sec;
    long tv_usec;
};
#endif

double dmax(double a, double b);
double dmin(double a, double b);
double dabs(double d);

double square(double x);

double ms_between(struct timeval *tv1, struct timeval *tv2);

double* compare_weights_vector;
int compare_weights(const void* v1, const void* v2);
int compare_weights_asc(const void* v1, const void* v2);

void permute_doubles(double* data, int* order, int N);
void permute_ints(int* data, int* order, int N);
void permute_nodes(Node** data, int* order, int N);

#ifdef WINDOWS
void gettimeofday(struct timeval* tv, void* null);
#endif

void setup_XYw(double* X, int NX, double* Y, int NY, int D, double* w, double x_low,
               double x_high, double y_low, double y_high, char* weights_mode);

int verification_check_sum(double* mginfls, double* naiveinfls, int NY, double eps_abs,
                        double eps_rel);

int verification_check_max(int* mginds, int* naiveinds, double* mginfls, double* naiveinfls, int NY);

void setup_kernel(char* name);

#endif

