MODE = "cl_cpu_by_feat"
## MODE = "cl_gpu_by_feat"

DATASET = "concrete"

PRECISION = "FLOAT"
PRECISION = "DOUBLE"

cat("MODE ", MODE, "  ")
cat("PRECISION ", PRECISION, "  ")

NPAR = 3

OMEGA = 1e-9

## ******************** DATASET
DATA = new.env()

if(DATASET == "concrete") {
  DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_X.txt")))
  DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_y.txt"))
}

## DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_X.txt")[1:1000,]))
## DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_y.txt"))[1:1000]

DATA$n = dim(DATA$X)[2]
DATA$d = dim(DATA$X)[1]

## ****************************************

PRIOR.SIGMA.SHAPE = 1
PRIOR.SIGMA.RATE = 0.1
PRIOR.LAMBDA.SHAPE = 1
PRIOR.LAMBDA.RATE = 0.1
PRIOR.TAU.SHAPE = 1
PRIOR.TAU.RATE = 0.1
PRIORS = c(PRIOR.SIGMA.SHAPE, PRIOR.SIGMA.RATE, PRIOR.LAMBDA.SHAPE, PRIOR.LAMBDA.RATE, PRIOR.TAU.SHAPE, PRIOR.TAU.RATE)

gradient = rep(0, NPAR)
rvect = rep(0, DATA$n)

if(MODE == "cl_cpu_by_feat") {
  if(PRECISION == "FLOAT") dyn.load("wrapper_gradient_cpu_float_by_feat.so")
  if(PRECISION == "DOUBLE") dyn.load("wrapper_gradient_cpu_double_by_feat.so")
}

if(MODE == "cl_gpu_by_feat") {
  if(PRECISION == "FLOAT") dyn.load("wrapper_gradient_gpu_float_by_feat.so")
  if(PRECISION == "DOUBLE") dyn.load("wrapper_gradient_gpu_double_by_feat.so")
}

THRS_CONVERGENCE_CG = 1.0e-8
THRS_EARLY_STOP_CG = sqrt(DATA$n) * 1.0
DECAY_EARLY_STOP_CG = 1.0

SEEDS = c(as.matrix(read.table("../test_linear_system/LIST_SEEDS.txt"))) ## 1421812 + 7 * c(1:100)
NSEEDS = length(SEEDS)
NREP = 100
plot.niterations = matrix(0, NSEEDS, NREP)
plot.condition = rep(0, NREP)
plot.norm2.error = matrix(0, NSEEDS, NREP)

for(index_seed in 1:NSEEDS) {
  cat("\n", index_seed, "\n")

  set.seed(SEEDS[index_seed])

  theta = log(rgamma(3, 0.3, 0.01))
  exp_theta = exp(theta)

  K = exp(theta[1]) * exp( - exp(theta[3]) * as.matrix(dist(t(DATA$X))^2)) + diag(exp(theta[2]) + 1e-9, DATA$n)
  K.inv = solve(K)
  lambda = eigen(K, symmetric=T, only.values=T)[[1]]
  plot.condition[index_seed] = lambda[1] / lambda[DATA$n]
  dKdt1 = exp(theta[1]) * exp( - exp(theta[3]) * as.matrix(dist(t(DATA$X))^2))
  dKdt2 = diag(exp_theta[2], DATA$n)
  dKdt3 = dKdt1 * (log(dKdt1) - theta[1])
  dKdt3[dKdt1 == 0] = 0
  
  grad_R = rep(0, 3)
  grad_R[1] = - 0.5 * sum(K.inv * dKdt1) + 0.5 * t(DATA$y) %*% K.inv %*% dKdt1 %*% K.inv %*% DATA$y + PRIOR.SIGMA.SHAPE - PRIOR.SIGMA.RATE * exp_theta[1];
  grad_R[2] = - 0.5 * sum(K.inv * dKdt2) + 0.5 * t(DATA$y) %*% K.inv %*% dKdt2 %*% K.inv %*% DATA$y + PRIOR.LAMBDA.SHAPE - PRIOR.LAMBDA.RATE * exp_theta[2];
  grad_R[3] = - 0.5 * sum(K.inv * dKdt3) + 0.5 * t(DATA$y) %*% K.inv %*% dKdt3 %*% K.inv %*% DATA$y + PRIOR.TAU.SHAPE - PRIOR.TAU.RATE * exp_theta[3];

  for(repetition in 1:NREP) {
    cat("\n", repetition, " ")
    
    niterations = 0
    res = .C("wrapper_gradient", as.double(DATA$X), as.double(DATA$y), as.integer(DATA$n), as.integer(DATA$d), as.double(theta), as.double(exp_theta), as.double(OMEGA), as.double(PRIORS), as.integer(NPAR), as.double(gradient), as.double(rvect), as.double(THRS_CONVERGENCE_CG), as.double(THRS_EARLY_STOP_CG), as.double(DECAY_EARLY_STOP_CG), as.integer(niterations), as.integer(SEEDS[index_seed]/2 + repetition))
    plot.niterations[index_seed, repetition] = res[[15]]
  
    ## rvect = res[[11]]
    ## grad_R = rep(0, 3)
    ## grad_R[1] = - 0.5 * t(rvect) %*% K.inv %*% dKdt1 %*% rvect + 0.5 * t(DATA$y) %*% K.inv %*% dKdt1 %*% K.inv %*% DATA$y + PRIOR.SIGMA.SHAPE - PRIOR.SIGMA.RATE * exp_theta[1];
    ## grad_R[2] = - 0.5 * t(rvect) %*% K.inv %*% dKdt2 %*% rvect + 0.5 * t(DATA$y) %*% K.inv %*% dKdt2 %*% K.inv %*% DATA$y + PRIOR.LAMBDA.SHAPE - PRIOR.LAMBDA.RATE * exp_theta[2];
    ## grad_R[3] = - 0.5 * t(rvect) %*% K.inv %*% dKdt3 %*% rvect + 0.5 * t(DATA$y) %*% K.inv %*% dKdt3 %*% K.inv %*% DATA$y + PRIOR.TAU.SHAPE - PRIOR.TAU.RATE * exp_theta[3];
    
    ## Total variance = sum variances for each element of the vector
    plot.norm2.error[index_seed, repetition] = sum((res[[10]] - grad_R)^2)
    
  }

  filesave = paste("RES_", DATASET, "_GRADIENT_GAMMA_", DECAY_EARLY_STOP_CG, "_THRS_OVER_SQRTN_", THRS_EARLY_STOP_CG / sqrt(DATA$n), "_", PRECISION, "_", SEEDS[index_seed], ".txt", sep="")
  
  cat(theta, plot.condition[index_seed], " ", mean(plot.niterations[index_seed, ]), " ", mean(plot.norm2.error[index_seed, ]), " ", sum(grad_R^2), "\n", file=filesave)
}
