######################################################################################################
######################################################################################################
################# Plot number of iterations and error vs condition number to estimate the gradient with and without early stop 
######################################################################################################
######################################################################################################

ps.options(width=10, height=10, paper="special", horizontal=F, pointsize=40)

DATASET = "concrete"

## SEEDS = 1421812 + 7 * c(1:100)
## SEEDS = SEEDS[1:50]
SEEDS = c(as.matrix(read.table("../test_linear_system/LIST_SEEDS.txt")))

PCH.LIST = c(3,4,22,23)

if(T) {

PRECISION = "DOUBLE"
  
GAMMA.THRS = matrix(0, 3, 2)
GAMMA.THRS[1,] = c(1, -0.01)
GAMMA.THRS[2,] = c(1, 0.1)
GAMMA.THRS[3,] = c(1, 1)

array_res = array(0, c(length(SEEDS), 7, dim(GAMMA.THRS)[1]))
for(index.method in 1:(dim(GAMMA.THRS)[1])) {

  GAMMA = GAMMA.THRS[index.method, 1]
  THRS = GAMMA.THRS[index.method, 2]
  
  for(i in 1:length(SEEDS)) {
    SEED = SEEDS[i]

    filename = paste("~/RES_ICML2015/gradient/RES_", DATASET, "_GRADIENT_GAMMA_", GAMMA, "_THRS_OVER_SQRTN_", THRS, "_", PRECISION, "_", SEED, ".txt", sep="")
    tmp = as.matrix(read.table(filename)[1,])
    array_res[i,,index.method] = tmp
  }
}

## par(mfrow=c(1,2))

## ************** N ITERATIONS vs CONDITION NUMBER
postscript("PLOT_GRADIENT_NITERATIONS_vs_CONDITION_NUMBER_GAMMA_1.eps")
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1]), log10(array_res[,5,1]), pch=PCH.LIST[1], ylim=range(log10(array_res[,5,])), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("number iterations")))
points(log10(array_res[,4,2]), log10(array_res[,5,2]), pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3]), log10(array_res[,5,3]), pch=PCH.LIST[3], col=3)
## points(log10(array_res[,4,4]), log10(array_res[,5,4]), pch=18, col=4)
legend(0, 3.88, c("CG", "ULISSE q=0.1", "ULISSE q=1.0"), pch=PCH.LIST, col=c(1:3), cex=0.8)
dev.off()

## ************** ERROR vs CONDITION NUMBER
postscript("PLOT_GRADIENT_ERROR_vs_CONDITION_NUMBER_GAMMA_1.eps")
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1]), log10(array_res[,6,1]) - log10(array_res[,7,1]), pch=PCH.LIST[1], ylim=c(range(log10(array_res[,6,]) - log10(array_res[,7,]))), xlim=c(range(log10(array_res[,4,]))), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("avg rel square norm")))
points(log10(array_res[,4,2]), log10(array_res[,6,2]) - log10(array_res[,7,2]), pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3]), log10(array_res[,6,3]) - log10(array_res[,7,3]), pch=PCH.LIST[3], col=3)
## points(log10(array_res[,4,4]), log10(array_res[,6,4]), pch=18, col=4)
## legend(2.5, -6, c("q=0.0", "q=0.1", "q=1.0"), pch=PCH.LIST, col=c(1:3), cex=0.85)
dev.off()

}
