#include "../test_linear_system/wrapper_linear_system.c"
#include "../test_gradient/dK_vector_multiplication.c"


void compute_gradient(double *X, double *y, int *n, int *d, double *kpar, double *exp_kpar, int *npar, double *OMEGA, gsl_rng *uniform_random_variable, double *priors, double *rvect, double *K_inv_y, double *K_inv_r, double *gradient, double THRS_CONVERGENCE_CG, double THRS_EARLY_STOP_CG, double DECAY_EARLY_STOP_CG, int *niterations) {

  int i;

  double tmp, tmp2, grad_part1 = 0, grad_part2 = 0, grad_part3;

  int niterations_tmp[2];

  /* // Draw a random vector r */
  /* for(i = 0; i < *n; i++) { */
  /* 	tmp = gsl_rng_uniform(uniform_random_variable); */
  /* 	if(tmp > 0.5) gsl_vector_set(rvect, i, 1); */
  /* 	else gsl_vector_set(rvect, i, -1); */
  /*  } */
  /* // ********************** */

  /* // Draw some random components of random vector r - with probability P = 1% redraw r[i] */
  /* for(i = 0; i < *n; i++) { */
  /* 	tmp2 = gsl_rng_uniform(uniform_random_variable); */
  /* 	if(tmp2 < 0.01) { */
  /* 	  tmp = gsl_rng_uniform(uniform_random_variable); */
  /* 	  if(tmp > 0.5) gsl_vector_set(rvect, i, 1); */
  /* 	  else gsl_vector_set(rvect, i, -1); */
  /* 	} */
  /*  } */
  /* // ********************** */

  // Compute an unbiased version of K^-1 r and two unbiased versions of K^-1 y
  niterations_tmp[0] = 0;
  niterations_tmp[1] = 0;
  solve_CG(X, n, d, kpar, exp_kpar, OMEGA, rvect, uniform_random_variable, 1, K_inv_r, THRS_CONVERGENCE_CG, THRS_EARLY_STOP_CG, DECAY_EARLY_STOP_CG, niterations_tmp);
  /* printf("-- %d %d -- ", niterations_tmp[0], niterations_tmp[1]); */

  *niterations += niterations_tmp[0];

  niterations_tmp[0] = 0;
  niterations_tmp[1] = 0;
  solve_CG(X, n, d, kpar, exp_kpar, OMEGA, y, uniform_random_variable, 2, K_inv_y, THRS_CONVERGENCE_CG, THRS_EARLY_STOP_CG, DECAY_EARLY_STOP_CG, niterations_tmp);
  /* printf("-- %d %d -- ", niterations_tmp[0], niterations_tmp[1]); */


  if(niterations_tmp[0] > niterations_tmp[1]) *niterations += niterations_tmp[0];
  else *niterations += niterations_tmp[1];

  double *K_inv_y_1 = &K_inv_y[0];
  double *K_inv_y_2 = &K_inv_y[*n];

  double *dK_r = (double*) malloc (*n * sizeof(double));
  double *dK_K_inv_y = (double*) malloc (*n * sizeof(double));
  
  // Compute dK r and dK K^-1 y for all parameters and use them to compute the components of the gradient
  for(i = 0; i < *npar; i++) {
	// Derivatives wrt psi_sigma and psi_tau on the GPU
	if(i != 1) {
	  dK_vector_multiplication(X, n, d, kpar, exp_kpar, OMEGA, rvect, i, dK_r);
	  dK_vector_multiplication(X, n, d, kpar, exp_kpar, OMEGA, K_inv_y_1, i, dK_K_inv_y);
	}
	
	// Derivative wrt psi_lambda
	else {
	  for(int j = 0; j < *n; j++) { dK_r[j] = exp_kpar[1] * rvect[j]; }
	  for(int j = 0; j < *n; j++) { dK_K_inv_y[j] = exp_kpar[1] * K_inv_y[j]; }
	}

  	grad_part1 = 0;
  	for(int j = 0; j < *n; j++) grad_part1 += dK_r[j] * K_inv_r[j];
  	grad_part1 *= -0.5;

  	grad_part2 = 0;
  	for(int j = 0; j < *n; j++) grad_part2 += dK_K_inv_y[j] * K_inv_y_2[j];
  	grad_part2 *= 0.5;

  	grad_part3 = 0;
  	grad_part3 = priors[0 + i * 2] - priors[1 + i * 2] * exp_kpar[i];
  	gradient[i] = grad_part1 + grad_part2 + grad_part3;
  }
}

int wrapper_gradient(double *DATA, double *y, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double *priors, int *npar, double *gradient, double *rvect, double *THRS_CONVERGENCE_CG_pointer, double *THRS_EARLY_STOP_CG_pointer, double *DECAY_EARLY_STOP_CG_pointer, int *niterations, int *SEED) {

  if(*d > MAX_N_DIM) { printf("PROBLEM DIMENSION IS TOO LARGE - CHANGE THE FUNCTION THAT IS EXECUTED ON THE GPU TO ALLOW FOR LARGER DIMENSIONS \n\n\n"); return 1; }
  
#if (HOST==HOST_DESKTOP)
  printf("---- COMPILED FOR DESKTOP ----");
#elif (HOST==HOST_CURIEUSE)
  printf("---- COMPILED FOR CURIEUSE ----");
#endif
  
#if (PRECISION == PREC_FLOAT)
  printf("---- FLOAT ----");
#elif (PRECISION == PREC_DOUBLE)
  printf("---- DOUBLE ----");
#endif
  
#if (DEVICE == DEV_CPU)
  printf("---- CL CPU ----");
#elif (DEVICE == DEV_GPU)
  printf("---- CL GPU ----");
#endif
  
#if (DATA_ORIENT == BY_DATA)
  printf("---- BY DATA ----");
#elif (DATA_ORIENT == BY_FEAT)
  printf("---- BY FEAT ----");
#endif

  int i, j;

  // Initialize random number generator
  const gsl_rng_type *type_random_sampling;
  gsl_rng *uniform_random_variable;

  type_random_sampling = gsl_rng_mt19937;
  uniform_random_variable = gsl_rng_alloc (type_random_sampling);
  gsl_rng_set(uniform_random_variable, *SEED);
  // **********************************

  // Initialize Bernoulli random vector r
  double tmp; 
  for(i = 0; i < *n; i++) {
    tmp = gsl_rng_uniform(uniform_random_variable);
    if(tmp > 0.5) rvect[i] = 1;
    else rvect[i] = -1;
  }
  double *K_inv_y = (double*) malloc (*n * 2 * sizeof(double));
  double *K_inv_r = (double*) malloc (*n * 1 * sizeof(double));

  for(i = 0; i < *n; i++) {
	K_inv_y[i] = 0;
	K_inv_y[i + *n] = 0;
	K_inv_r[i] = 0;
  }

  compute_gradient(DATA, y, n, d, kpar, exp_kpar, npar, OMEGA, uniform_random_variable, priors, rvect, K_inv_y, K_inv_r, gradient, *THRS_CONVERGENCE_CG_pointer, *THRS_EARLY_STOP_CG_pointer, *DECAY_EARLY_STOP_CG_pointer, niterations);

  return 0;
}
