#define SPLIT_ROW_MATRIX 8

#define MAX_N_DIM 100

#define HOST_DESKTOP 0
#define HOST_CURIEUSE 1

#define PREC_FLOAT 0
#define PREC_DOUBLE 1

#define DEV_CPU 0
#define DEV_GPU 1

#define BY_DATA 0
#define BY_FEAT 1

#include <time.h>
#include <stdio.h>
#include <stdlib.h>

#if (PRECISION == PREC_FLOAT)
  #if (DATA_ORIENT == BY_DATA)
    #define PROGRAM_FILE "gpu_kernel_matrix_vector_multiplication_float_by_data.cl"
  #elif (DATA_ORIENT == BY_FEAT)
    #define PROGRAM_FILE "gpu_kernel_matrix_vector_multiplication_float_by_feat.cl"
  #endif
#elif (PRECISION == PREC_DOUBLE)
  #if (DATA_ORIENT == BY_DATA)
    #define PROGRAM_FILE "gpu_kernel_matrix_vector_multiplication_double_by_data.cl"
  #elif (DATA_ORIENT == BY_FEAT)
    #define PROGRAM_FILE "gpu_kernel_matrix_vector_multiplication_double_by_feat.cl"
  #endif
#endif

#if (HOST==HOST_DESKTOP)
  #include <OpenCL/opencl.h>
#elif (HOST==HOST_CURIEUSE)
  #include <CL/opencl.h>
#endif


#include "gpu_utilities.c"
  
int wrapper_kernel_matrix_vector_multiplication(double *X, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double* v, int *type, double *y) {

  if(*d > MAX_N_DIM) { printf("PROBLEM DIMENSION IS TOO LARGE - CHANGE THE FUNCTION THAT IS EXECUTED ON THE GPU TO ALLOW FOR LARGER DIMENSIONS \n\n\n"); return 1; }
  
#if (HOST==HOST_DESKTOP)
  printf("---- COMPILED FOR DESKTOP ----");
#elif (HOST==HOST_CURIEUSE)
  printf("---- COMPILED FOR CURIEUSE ----");
#endif
  
#if (PRECISION == PREC_FLOAT)
  printf("---- FLOAT ----");
#elif (PRECISION == PREC_DOUBLE)
  printf("---- DOUBLE ----");
#endif
  
#if (DEVICE == DEV_CPU)
  printf("---- CL CPU ----");
#elif (DEVICE == DEV_GPU)
  printf("---- CL GPU ----");
#endif
  
#if (DATA_ORIENT == BY_DATA)
  printf("---- BY DATA ----");
#elif (DATA_ORIENT == BY_FEAT)
  printf("---- BY FEAT ----");
#endif
  
  int i, j;

#if (PRECISION == PREC_FLOAT)
  
  float *X_float; X_float = (float*) malloc (*n * *d * sizeof(float));
  float *kpar_float; kpar_float = (float*) malloc (3 * sizeof(float));
  float *exp_kpar_float; exp_kpar_float = (float*) malloc (3 * sizeof(float));
  float *OMEGA_float; OMEGA_float = (float*) malloc (sizeof(float));
  float *v_float; v_float = (float*) malloc (*n * sizeof(float));

  float *res_float[N_DEVICES];  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    res_float[index_devices] = (float*) malloc (*n * SPLIT_ROW_MATRIX * sizeof(float));

#if (DATA_ORIENT == BY_DATA)
  for(i = 0; i < *n; i++) {
    for(j = 0; j < *d; j++) {
      X_float[i + j * *n] = (float) X[i + j * *n];
    }
  }
#elif (DATA_ORIENT == BY_FEAT)
  for(i = 0; i < *d; i++) {
    for(j = 0; j < *n; j++) {
      X_float[i + j * *d] = (float) X[i + j * *d];
    }
  }
#endif
  
  for(i = 0; i < 3; i++) {
    kpar_float[i] = (float) kpar[i];
    exp_kpar_float[i] = (float) exp_kpar[i];
  }

  OMEGA_float[0] = (float) OMEGA[0];
  
  for(i = 0; i < *n; i++) {
    v_float[i] = (float) v[i];
  }

#elif (PRECISION == PREC_DOUBLE)

  double *X_double; X_double = X;
  double *kpar_double; kpar_double = kpar;
  double *exp_kpar_double; exp_kpar_double = exp_kpar;
  double *OMEGA_double; OMEGA_double = OMEGA;
  double *v_double; v_double = v;

  double *res_double[N_DEVICES];  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    res_double[index_devices] = (double*) malloc (*n * SPLIT_ROW_MATRIX * sizeof(double));

#endif

  cl_context context[N_DEVICES]; cl_program program[N_DEVICES]; cl_kernel kernel[N_DEVICES]; cl_command_queue command_queue[N_DEVICES]; 
  cl_uint num_of_platforms=0; cl_platform_id platform_id; cl_device_id device_id[N_DEVICES]; cl_uint num_of_devices=0; cl_int err; 

  cl_mem X_gpu[N_DEVICES], n_gpu[N_DEVICES], d_gpu[N_DEVICES], kpar_gpu[N_DEVICES], exp_kpar_gpu[N_DEVICES], OMEGA_gpu[N_DEVICES], v_gpu[N_DEVICES];
  cl_mem split_row_matrix_gpu[N_DEVICES], n_devices_gpu[N_DEVICES], gpu_id_gpu[N_DEVICES], res_gpu[N_DEVICES];

  int *split_row_matrix_pointer[N_DEVICES], *n_devices_pointer[N_DEVICES], *gpu_id_pointer[N_DEVICES];
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
    split_row_matrix_pointer[index_devices] = (int*) malloc (sizeof(int)); split_row_matrix_pointer[index_devices][0] = SPLIT_ROW_MATRIX;
    n_devices_pointer[index_devices] = (int*) malloc (sizeof(int)); n_devices_pointer[index_devices][0] = N_DEVICES;
    gpu_id_pointer[index_devices] = (int*) malloc (sizeof(int)); gpu_id_pointer[index_devices][0] = index_devices;
  }
  
  // retreive a list of platforms avaible
  if (clGetPlatformIDs(1, &platform_id, &num_of_platforms)!= CL_SUCCESS) { printf("Unable to get platform_id\n"); return 1; }
  
  // try to get a supported GPU device
#if (DEVICE == DEV_CPU)
  if (clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_CPU, N_DEVICES, device_id, &num_of_devices) != CL_SUCCESS) { printf("Unable to get device_id\n"); return 1; }
#elif (DEVICE == DEV_GPU)
  if (clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, N_DEVICES, device_id, &num_of_devices) != CL_SUCCESS) { printf("Unable to get device_id\n"); return 1; }
#endif

  printf("Detected %d devices - Using %d devices\n\n", num_of_devices, N_DEVICES);
  
  // context properties list - must be terminated with 0
  cl_context_properties properties[3]; properties[0]= CL_CONTEXT_PLATFORM; properties[1]= (cl_context_properties) platform_id; properties[2]= 0;
  
  // create a context with the GPU device
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    context[index_devices] = clCreateContext(properties, N_DEVICES, device_id, NULL, NULL, &err);  

  // create command queue using the context and device
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    command_queue[index_devices] = clCreateCommandQueue(context[index_devices], device_id[index_devices], 0, &err);
  
  // create a program from the kernel source code
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    program[index_devices] = build_program(context[index_devices], device_id[index_devices], PROGRAM_FILE);

  // create buffers for the input and ouput
#if (PRECISION == PREC_FLOAT)
  // specify which kernel from the program to execute
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    kernel[index_devices] = clCreateKernel(program[index_devices], "matrix_vector_multiplication_float", &err);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    X_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * *n * *d, NULL, NULL);
    n_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    d_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * 3, NULL, NULL);
    exp_kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * 3, NULL, NULL);
    OMEGA_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float), NULL, NULL);
    v_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * *n, NULL, NULL);
    split_row_matrix_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    n_devices_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    gpu_id_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    res_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_WRITE_ONLY, sizeof(float) * *n * SPLIT_ROW_MATRIX, NULL, NULL);
  }

#elif (PRECISION == PREC_DOUBLE)
  // specify which kernel from the program to execute
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    kernel[index_devices] = clCreateKernel(program[index_devices], "matrix_vector_multiplication_double", &err);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    X_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * *n * *d, NULL, NULL);
    n_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    d_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * 3, NULL, NULL);
    exp_kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * 3, NULL, NULL);
    OMEGA_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double), NULL, NULL);
    v_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * *n, NULL, NULL);
    split_row_matrix_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    n_devices_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    gpu_id_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    res_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_WRITE_ONLY, sizeof(double) * *n * SPLIT_ROW_MATRIX, NULL, NULL);
  }

#endif
  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
  // set the argument list for the kernel command
    clSetKernelArg(kernel[index_devices], 0, sizeof(cl_mem), &X_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 1, sizeof(cl_mem), &n_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 2, sizeof(cl_mem), &d_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 3, sizeof(cl_mem), &kpar_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 4, sizeof(cl_mem), &exp_kpar_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 5, sizeof(cl_mem), &OMEGA_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 6, sizeof(cl_mem), &v_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 7, sizeof(cl_mem), &split_row_matrix_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 8, sizeof(cl_mem), &n_devices_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 9, sizeof(cl_mem), &gpu_id_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 10, sizeof(cl_mem), &res_gpu[index_devices]);
  }

  for(int jj = 0; jj < 10; jj++) {

  // load data into the input buffer
#if (PRECISION == PREC_FLOAT)

    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      clEnqueueWriteBuffer(command_queue[index_devices], X_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n * *d, X_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_gpu[index_devices], CL_TRUE, 0, sizeof(int), n, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], d_gpu[index_devices], CL_TRUE, 0, sizeof(int), d, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], kpar_gpu[index_devices], CL_TRUE, 0, sizeof(float) * 3, kpar_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], exp_kpar_gpu[index_devices], CL_TRUE, 0, sizeof(float) * 3, exp_kpar_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], OMEGA_gpu[index_devices], CL_TRUE, 0, sizeof(float), OMEGA_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n, v_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], split_row_matrix_gpu[index_devices], CL_TRUE, 0, sizeof(int), split_row_matrix_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_devices_gpu[index_devices], CL_TRUE, 0, sizeof(int), n_devices_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], gpu_id_gpu[index_devices], CL_TRUE, 0, sizeof(int), gpu_id_pointer[index_devices], 0, NULL, NULL);
    }

#elif (PRECISION == PREC_DOUBLE)

    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      clEnqueueWriteBuffer(command_queue[index_devices], X_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n * *d, X_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_gpu[index_devices], CL_TRUE, 0, sizeof(int), n, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], d_gpu[index_devices], CL_TRUE, 0, sizeof(int), d, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, kpar_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], exp_kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, exp_kpar_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], OMEGA_gpu[index_devices], CL_TRUE, 0, sizeof(double), OMEGA_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n, v_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], split_row_matrix_gpu[index_devices], CL_TRUE, 0, sizeof(int), split_row_matrix_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_devices_gpu[index_devices], CL_TRUE, 0, sizeof(int), n_devices_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], gpu_id_gpu[index_devices], CL_TRUE, 0, sizeof(int), gpu_id_pointer[index_devices], 0, NULL, NULL);
    }

#endif
  
  size_t global[2];
  global[0] = *n;
  global[1] = SPLIT_ROW_MATRIX;

  // enqueue the kernel command for execution
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueNDRangeKernel(command_queue[index_devices], kernel[index_devices], 2, NULL, &global[0], NULL, 0, NULL, NULL);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clFinish(command_queue[index_devices]);

#if (PRECISION == PREC_FLOAT)
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n * SPLIT_ROW_MATRIX, res_float[index_devices], 0, NULL, NULL);
#elif (PRECISION == PREC_DOUBLE)
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n * SPLIT_ROW_MATRIX, res_double[index_devices], 0, NULL, NULL);
#endif

  }

#if (PRECISION == PREC_FLOAT)

  for(i = 0; i < *n; i++) {
    /* printf("%lf %lf \n",res_float[i], res_float[i + *n]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
	y[i] += res_float[index_devices][i + j * *n];
      }
    }
    y[i] += (exp_kpar_float[1] + *OMEGA_float) * v_float[i];
  }

#elif (PRECISION == PREC_DOUBLE)

  for(i = 0; i < *n; i++) {
    /* printf("%lf %lf \n",res_double[i], res_double[i + *n]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
	y[i] += res_double[index_devices][i + j * *n];
      }
    }
    y[i] += (exp_kpar[1] + *OMEGA) * v_double[i];
  }

#endif

  /* printf("\n\n"); */

  // cleanup - release OpenCL resources
  
#if (PRECISION == PREC_FLOAT)

  free(X_float);
  free(kpar_float);
  free(exp_kpar_float);
  free(OMEGA_float);
  free(v_float);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    free(res_float[index_devices]);
  }

#elif (PRECISION == PREC_DOUBLE)

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    free(res_double[index_devices]);
  }

#endif

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    clReleaseMemObject(X_gpu[index_devices]);
 
    clReleaseMemObject(d_gpu[index_devices]);
    clReleaseMemObject(n_gpu[index_devices]);
    clReleaseMemObject(kpar_gpu[index_devices]);
    clReleaseMemObject(exp_kpar_gpu[index_devices]);
    clReleaseMemObject(OMEGA_gpu[index_devices]);
    clReleaseMemObject(v_gpu[index_devices]);
    clReleaseMemObject(split_row_matrix_gpu[index_devices]);
    clReleaseMemObject(n_devices_gpu[index_devices]);
    clReleaseMemObject(gpu_id_gpu[index_devices]);
    clReleaseMemObject(res_gpu[index_devices]);

    clReleaseProgram(program[index_devices]);
    clReleaseKernel(kernel[index_devices]);
    clReleaseCommandQueue(command_queue[index_devices]);
    clReleaseContext(context[index_devices]);
  }

  
  return 0;
}
