conjgrad.preconditioned = function(b, delta, tol = 1e-8) {

  n = DATA$n
  d = DATA$d
  
  NITERATIONS <<- 0
  NITERATIONS.INNER <<- 0
  DELTA.PRECOND = OMEGA + delta
  
  x0 = x = rep(0, length(b))
  r = b ## - A %*% x

  res = .C("wrapper_linear_system", as.double(DATA$X), as.integer(DATA$n), as.integer(DATA$d), as.double(theta), as.double(exp_theta), as.double(DELTA.PRECOND), as.double(r), as.integer(n_results), as.double(result), as.double(THRS_CONVERGENCE_CG), as.double(THRS_EARLY_STOP_CG), as.double(DECAY_EARLY_STOP_CG), as.integer(niterations))
  ## z = M.inv %*% r
  result = z = res[[9]]
  NITERATIONS <<- NITERATIONS + res[[13]]
  NITERATIONS.INNER <<- NITERATIONS.INNER + res[[13]]

  p = z
  rsold = as.double(t(r) %*% z)

  for (k in 1:(2*n)) {
    res = .C("wrapper_kernel_matrix_vector_multiplication", as.double(DATA$X), as.integer(DATA$n), as.integer(DATA$d), as.double(theta), as.double(exp_theta), as.double(OMEGA), as.double(p), as.integer(TYPE_KERNEL_MATRIX_VECTOR_MULTIPLICATION), as.double(result))
    ## res = .C("wrapper_multiply_K_vector", as.double(DATA$X), as.integer(DATA$n), as.integer(DATA$d), as.double(theta), as.double(exp_theta), as.double(OMEGA), as.double(p), as.integer(TYPE_KERNEL_MATRIX_VECTOR_MULTIPLICATION), as.double(result))
    
    NITERATIONS <<- NITERATIONS + 1
    Ap = res[[9]]
    ## Ap = A %*% p
   
    alpha = rsold / as.double((t(p) %*% Ap))

    x = x + alpha * p
    r = r - alpha * Ap

    res = .C("wrapper_linear_system", as.double(DATA$X), as.integer(DATA$n), as.integer(DATA$d), as.double(theta), as.double(exp_theta), as.double(DELTA.PRECOND), as.double(r), as.integer(n_results), as.double(result), as.double(THRS_CONVERGENCE_CG), as.double(THRS_EARLY_STOP_CG), as.double(DECAY_EARLY_STOP_CG), as.integer(niterations))
    ## z = M.inv %*% r
    result = z = res[[9]]
    NITERATIONS <<- NITERATIONS + res[[13]]
    NITERATIONS.INNER <<- NITERATIONS.INNER + res[[13]]
        
    rsnew = as.double(t(r) %*% z)
    if (sqrt(crossprod(r)) < tol) break

    p = z + rsnew / rsold * p

    rsold = rsnew
  }

  cat(NITERATIONS, "\n")

  x
}

for(OPTION_DELTA in 10^(c(-1, -2, -3, -4, -5, -6, -7, -8))) {

MODE = "cl_cpu_by_feat"
## MODE = "cl_gpu_by_feat"

DATASET = "concrete"

PRECISION = "FLOAT"
PRECISION = "DOUBLE"

cat("MODE ", MODE, "  ")
cat("PRECISION ", PRECISION, "  ")

## PERFORM THE MATRIX-VECTOR PRODUCT IN THE CG ALGORITHM USING DOUBLE PRECISION
if(MODE == "cl_cpu_by_feat") {
  ## if(PRECISION == "FLOAT") dyn.load("../test_kernel_matrix_vector_multiplication/gpu_wrapper_cpu_float_by_feat.so")
  ## if(PRECISION == "DOUBLE")
  dyn.load("../test_kernel_matrix_vector_multiplication/gpu_wrapper_cpu_double_by_feat.so")
}

if(MODE == "cl_gpu_by_feat") {
  ## if(PRECISION == "FLOAT") dyn.load("../test_kernel_matrix_vector_multiplication/gpu_wrapper_gpu_float_by_feat.so")
  ## if(PRECISION == "DOUBLE")
  dyn.load("../test_kernel_matrix_vector_multiplication/gpu_wrapper_gpu_double_by_feat.so")
}

## dyn.load("../test_kernel_matrix_vector_multiplication/kernel_matrix_multiplication.so")
TYPE_KERNEL_MATRIX_VECTOR_MULTIPLICATION = 4 ## USELESS FOR MATRIX-VECTOR MULTIPLICATION CODE ON THE GPU BUT NEEDED IN THE CALL





## ******************** DATASET
DATA = new.env()
if(DATASET == "concrete") {
  DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_X.txt")))
  DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_y.txt"))
}

## DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_X.txt")[1:1000,]))
## DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_y.txt"))[1:1000]

DATA$n = dim(DATA$X)[2]
DATA$d = dim(DATA$X)[1]

## ****************************************

THRS_CONVERGENCE_CG = 1.0e-8
THRS_EARLY_STOP_CG = -sqrt(DATA$n) * 0.01
DECAY_EARLY_STOP_CG = 10.0

## n = 1000
## d = 4
## X = matrix(rnorm(d*n), d, n)

n_results = 1
result = matrix(0, DATA$n, n_results)
niterations = 0

  if(MODE == "cl_cpu_by_feat") {
    if(PRECISION == "FLOAT") dyn.load("wrapper_linear_system_cpu_float_by_feat.so")
    if(PRECISION == "DOUBLE") dyn.load("wrapper_linear_system_cpu_double_by_feat.so")
  }

  if(MODE == "cl_gpu_by_feat") {
    if(PRECISION == "FLOAT") dyn.load("wrapper_linear_system_gpu_float_by_feat.so")
    if(PRECISION == "DOUBLE") dyn.load("wrapper_linear_system_gpu_double_by_feat.so")
  }

OMEGA = 1e-9

SEEDS = c(as.matrix(read.table("LIST_SEEDS.txt"))) ## 1421812 + 7 * c(1:100)
NREP = length(SEEDS)
plot.niterations1 = plot.niterations2 = rep(0, NREP)
plot.condition = rep(0, NREP)
plot.norm2.error = rep(0, NREP)

for(repetition in 1:NREP) {
  cat("\n", repetition, "\n")

  set.seed(SEEDS[repetition])

  theta = log(rgamma(3, 0.3, 0.01))
  exp_theta = exp(theta)

  K = exp(theta[1]) * exp( - exp(theta[3]) * as.matrix(dist(t(DATA$X))^2)) + diag(exp(theta[2]) + 1e-9, DATA$n)
  lambda = eigen(K, symmetric=T, only.values=T)[[1]]
  plot.condition[repetition] = lambda[1] / lambda[DATA$n]


  res_linear_sytstem = conjgrad.preconditioned(DATA$y, OPTION_DELTA) 
  plot.niterations1[repetition] = NITERATIONS

  L = t(chol(K))
  plot.norm2.error[repetition] = sum((res_linear_sytstem - backsolve(t(L), forwardsolve(L, DATA$y)))^2)
  

  filesave = paste("RES_", DATASET, "_PRECONDITIONED_", OPTION_DELTA, "_", PRECISION, "_", SEEDS[repetition], ".txt", sep="")
  
  cat(theta, plot.condition[repetition], " ", plot.niterations1[repetition], NITERATIONS.INNER, " ", plot.norm2.error[repetition], "\n", file=filesave)
}

}

par(mfrow = c(1,2))
plot(log10(plot.condition), log10(plot.niterations1), pch=20)
plot(log10(plot.condition), log10(plot.norm2.error), pch=20)
