MODE = "cl_cpu_by_feat"
## MODE = "cl_gpu_by_feat"

DATASET = "concrete"

for(PRECISION in c("FLOAT", "DOUBLE")) {
## PRECISION = "FLOAT"
## PRECISION = "DOUBLE"

cat("MODE ", MODE, "  ")
cat("PRECISION ", PRECISION, "  ")



## ******************** DATASET
DATA = new.env()

if(DATASET == "concrete") {
  DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_X.txt")))
  DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_y.txt"))
}

## DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_X.txt")[1:1000,]))
## DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_y.txt"))[1:1000]

DATA$n = dim(DATA$X)[2]
DATA$d = dim(DATA$X)[1]

## ****************************************

THRS_CONVERGENCE_CG = 1.0e-8
THRS_EARLY_STOP_CG = -sqrt(DATA$n) * 0.1
DECAY_EARLY_STOP_CG = 10.0

## n = 1000
## d = 4
## X = matrix(rnorm(d*n), d, n)

n_results = 1
result = matrix(0, DATA$n, n_results)
niterations = 0

  if(MODE == "cl_cpu_by_feat") {
    if(PRECISION == "FLOAT") dyn.load("wrapper_linear_system_cpu_float_by_feat.so")
    if(PRECISION == "DOUBLE") dyn.load("wrapper_linear_system_cpu_double_by_feat.so")
  }

  if(MODE == "cl_gpu_by_feat") {
    if(PRECISION == "FLOAT") dyn.load("wrapper_linear_system_gpu_float_by_feat.so")
    if(PRECISION == "DOUBLE") dyn.load("wrapper_linear_system_gpu_double_by_feat.so")
  }

OMEGA = 1e-9

SEEDS = c(as.matrix(read.table("LIST_SEEDS.txt"))) ## 1421812 + 7 * c(1:100)
NREP = length(SEEDS)
plot.niterations1 = plot.niterations2 = rep(0, NREP)
plot.condition = rep(0, NREP)
plot.norm2.error = rep(0, NREP)

for(repetition in 1:NREP) {
  cat("\n", repetition, "\n")
  
  set.seed(SEEDS[repetition])

  theta = log(rgamma(3, 0.3, 0.01))
  exp_theta = exp(theta)

  K = exp(theta[1]) * exp( - exp(theta[3]) * as.matrix(dist(t(DATA$X))^2)) + diag(exp(theta[2]) + 1e-9, DATA$n)
  lambda = eigen(K, symmetric=T, only.values=T)[[1]]
  plot.condition[repetition] = lambda[1] / lambda[DATA$n]
  
  res = .C("wrapper_linear_system", as.double(DATA$X), as.integer(DATA$n), as.integer(DATA$d), as.double(theta), as.double(exp_theta), as.double(OMEGA), as.double(DATA$y), as.integer(n_results), as.double(result), as.double(THRS_CONVERGENCE_CG), as.double(THRS_EARLY_STOP_CG), as.double(DECAY_EARLY_STOP_CG), as.integer(niterations))
  plot.niterations1[repetition] = res[[13]]

  L = t(chol(K))
  plot.norm2.error[repetition] = sum((res[[9]] - backsolve(t(L), forwardsolve(L, DATA$y)))^2)
  
  ## rvect = sample(c(-1,1), DATA$n, replace=T)
  ## res = .C("wrapper_linear_system", as.double(DATA$X), as.integer(DATA$n), as.integer(DATA$d), as.double(theta), as.double(exp_theta), as.double(OMEGA), as.double(rvect), as.integer(n_results), as.double(result), as.double(THRS_CONVERGENCE_CG), as.double(THRS_EARLY_STOP_CG), as.double(DECAY_EARLY_STOP_CG), as.integer(niterations))
  ## plot.niterations2[repetition] = res[[13]]

## cat("Condition number: ", condition.number, "\n\n")

  filesave = paste("RES_", DATASET, "_", PRECISION, "_", SEEDS[repetition], ".txt", sep="")
  
  cat(theta, plot.condition[repetition], " ", plot.niterations1[repetition], " ", plot.norm2.error[repetition], "\n", file=filesave)
}


}

par(mfrow = c(1,2))
plot(log10(plot.condition), log10(plot.niterations1), pch=20)
plot(log10(plot.condition), log10(plot.norm2.error), pch=20)
