######################################################################################################
######################################################################################################
################# Plot number of iterations and error vs condition number for DOUBLE, FLOAT, and multipole methods
######################################################################################################
######################################################################################################

ps.options(width=10, height=10, paper="special", horizontal=F, pointsize=40)

## SEEDS = 1421812 + 7 * c(1:100)
## SEEDS = SEEDS[1:50]
SEEDS = c(as.matrix(read.table("LIST_SEEDS.txt")))

PCH.LIST = c(3,4,22,23)


if(F) {

METHODS = c("DOUBLE", "FLOAT", "ANCHORS", "KDTREE")

array_res = array(0, c(length(SEEDS), 6, length(METHODS)))
for(index.method in 1:length(METHODS)) {

  METHOD = METHODS[index.method]
  
  for(i in 1:length(SEEDS)) {
    SEED = SEEDS[i]

    filename = paste("~/RES_ICML2015/linear_system/RES_concrete_", METHOD, "_", SEED, ".txt", sep="")
    tmp = as.matrix(read.table(filename)[1,])
    array_res[i,,index.method] = tmp
  }
}

## par(mfrow=c(1,2))

## ************** N ITERATIONS vs CONDITION NUMBER
postscript("PLOT_NITERATIONS_vs_CONDITION_NUMBER.eps")
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1]), log10(array_res[,5,1]), pch=PCH.LIST[1], ylim=range(log10(array_res[,5,])), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("number iterations")))
points(log10(array_res[,4,2]), log10(array_res[,5,2]), pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3]), log10(array_res[,5,3]), pch=PCH.LIST[3], col=3)
points(log10(array_res[,4,4]), log10(array_res[,5,4]), pch=PCH.LIST[4], col=4)
legend(0, 5, c("double", "float", "anchors", "kdtree"), pch=PCH.LIST, col=c(1:4), cex=0.8)
dev.off()

## ************** N ITERATIONS RELATIVE TO DOUBLE vs CONDITION NUMBER
## plot(log10(array_res[,4,2]), array_res[,5,2] / array_res[,5,1], pch=16, col=2, ylim=c(0.5, 4.8))
## points(log10(array_res[,4,3]), array_res[,5,3] / array_res[,5,1], pch=17, col=3)
## points(log10(array_res[,4,4]), array_res[,5,4] / array_res[,5,1], pch=18, col=4)
## abline(h=1, lty=2)

## ************** ERROR vs CONDITION NUMBER - ERROR is the norm of the difference between the solutions computed with CG algorithm and Cholesky
postscript("PLOT_ERROR_vs_CONDITION_NUMBER.eps")
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1]), log10(array_res[,6,1])/2, pch=PCH.LIST[1], ylim=c(range(log10(array_res[,6,])/2)), xlim=c(range(log10(array_res[,4,]))), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("norm residual")))
points(log10(array_res[,4,2]), log10(array_res[,6,2])/2, pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3]), log10(array_res[,6,3])/2, pch=PCH.LIST[3], col=3)
points(log10(array_res[,4,4]), log10(array_res[,6,4])/2, pch=PCH.LIST[4], col=4)
legend(4.25, -18.7, c("double", "float", "anchors", "kdtree"), pch=PCH.LIST, col=c(1:4), cex=0.8)
dev.off()

## ************** ERROR RELTIVE TO DOUBLE vs CONDITION NUMBER
## plot(log10(array_res[,4,2]), log10(array_res[,6,2]) - log10(array_res[,6,1]), pch=16, xlim=c(range(log10(array_res[,4,]))), col=2, ylim=c(0,20))
## points(log10(array_res[,4,3]), log10(array_res[,6,3]) - log10(array_res[,6,1]), pch=17, col=3)
## points(log10(array_res[,4,4]), log10(array_res[,6,4]) - log10(array_res[,6,1]), pch=18, col=4)



## ************** N ITERATIONS vs ERROR
## plot(log10(array_res[,6,1]), log10(array_res[,5,1]), pch=20, ylim=c(0,5), xlim=c(range(log10(array_res[,6,]))))
## points(log10(array_res[,6,2]), log10(array_res[,5,2]), pch=20, col=2)
## points(log10(array_res[,6,3]), log10(array_res[,5,3]), pch=20, col=3)
## points(log10(array_res[,6,4]), log10(array_res[,5,4]), pch=20, col=4)



## ## ************** BOXPLOT - N ITERATIONS vs CONDITION NUMBER
## BATCH.SIZE = 10
## tmp = sort(log10(array_res[,4,1]), index.ret=T)
## sorted.condition.number = matrix(log10(array_res[,4,1])[tmp[[2]]], nrow=BATCH.SIZE)
## sorted.n.iterations = matrix(log10(array_res[,5,1])[tmp[[2]]], nrow=BATCH.SIZE)


## ## Prepare the frame for the boxplot
## plot(-999,-999, xlim=range(log10(array_res[,4,1])), ylim=range(log10(array_res[,5,])), xlab="log10(Condition Number)", ylab="log10(N Iterations)")
## boxplot(sorted.n.iterations, at=apply(sorted.condition.number, 2, mean), boxwex=0.05, add=T, axes=F, range=0)

## sorted.n.iterations = matrix(log10(array_res[,5,2])[tmp[[2]]], nrow=BATCH.SIZE)
## boxplot(sorted.n.iterations, at=apply(sorted.condition.number, 2, mean), boxwex=0.05, add=T, axes=F, range=0, col=2)

## sorted.n.iterations = matrix(log10(array_res[,5,3])[tmp[[2]]], nrow=BATCH.SIZE)
## boxplot(sorted.n.iterations, at=apply(sorted.condition.number, 2, mean), boxwex=0.05, add=T, axes=F, range=0, col=3)


}






######################################################################################################
######################################################################################################
################# Plot number of iterations and error vs condition number for preconditioning
######################################################################################################
######################################################################################################

if(F) {

METHODS = c("DOUBLE", "FLOAT")

array_res = array(0, c(length(SEEDS), 7, 3))
for(i in 1:length(SEEDS)) {
  SEED = SEEDS[i]

  filename = paste("~/RES_ICML2015/linear_system/RES_concrete_DOUBLE_", SEED, ".txt", sep="")
  tmp = as.matrix(read.table(filename)[1,])
  array_res[i,c(1:5, 7),1] = tmp
}

for(index.method in 1:length(METHODS)) {

  METHOD = METHODS[index.method]
  
  if(METHOD == "DOUBLE") DELTA = 1e-6
  if(METHOD == "FLOAT") DELTA = 1e-3
  
  for(i in 1:length(SEEDS)) {
    SEED = SEEDS[i]

    filename = paste("~/RES_ICML2015/linear_system/RES_concrete_PRECONDITIONED_", DELTA, "_", METHOD, "_", SEED, ".txt", sep="")
    tmp = as.matrix(read.table(filename)[1,])
    array_res[i,,index.method+1] = tmp
}




}

## par(mfrow=c(1,2))

## ************** N ITERATIONS vs CONDITION NUMBER
postscript(paste("PLOT_PRECONDITIONER_NITERATIONS_vs_CONDITION_NUMBER.eps", sep=""))
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1]), log10(array_res[,5,1]), pch=PCH.LIST[1], ylim=range(log10(array_res[,5,])), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("number iterations")))
points(log10(array_res[,4,2]), log10(array_res[,5,2]), pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3]), log10(array_res[,5,3]), pch=PCH.LIST[3], col=3)
legend(0, max(log10(array_res[,5,])), c("CG", "PCG double", "PCG float"), pch=PCH.LIST[1], col=c(1:3), cex=0.8)
dev.off()

## ************** ERROR vs CONDITION NUMBER
postscript(paste("PLOT_PRECONDITIONER_ERROR_vs_CONDITION_NUMBER.eps", sep=""))
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1]), log10(array_res[,7,1])/2, pch=PCH.LIST[1], ylim=c(range(log10(array_res[,7,])/2)), xlim=c(range(log10(array_res[,4,]))), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("norm residual")))
points(log10(array_res[,4,2]), log10(array_res[,7,2])/2, pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3]), log10(array_res[,7,3])/2, pch=PCH.LIST[3], col=3)
## legend(3.9, -24, c("CG", "PCG double", "PCG float"), pch=PCH.LIST, col=c(1:3), cex=0.8)
dev.off()

}



######################################################################################################
######################################################################################################
################# Plot variance early stop
######################################################################################################
######################################################################################################

if(F) {

THRSS = c(0.1, 1)
GAMMAS = c(100)

array_res = array(0, c(length(SEEDS), 6, length(THRSS)+1, length(GAMMAS)))
for(i in 1:length(SEEDS)) {
  SEED = SEEDS[i]

  filename = paste("~/RES_ICML2015/linear_system/RES_concrete_DOUBLE_", SEED, ".txt", sep="")
  tmp = as.matrix(read.table(filename)[1,])
  array_res[i,,1,1] = tmp
}

for(index.thrs in 1:length(THRSS)) {

  for(index.gamma in 1:length(GAMMAS)) {

    THRS = THRSS[index.thrs]
    GAMMA = GAMMAS[index.gamma]
  
    for(i in 1:length(SEEDS)) {
      SEED = SEEDS[i]
      
      filename = paste("~/RES_ICML2015/linear_system/RES_concrete_EARLYSTOP_DECAY_", GAMMA, "_THRS_OVER_SQRTN_", THRS, "_DOUBLE_", SEED, ".txt", sep="")
      tmp = as.matrix(read.table(filename)[1,])
      array_res[i,,index.thrs+1, index.gamma] = tmp
    }
  }
}

## par(mfrow=c(2,2)) 

## ************** N ITERATIONS vs CONDITION NUMBER
postscript(paste("PLOT_EARLYSTOP_GAMMA_", GAMMA,"_NITERATIONS_vs_CONDITION_NUMBER.eps", sep=""))
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1,1]), log10(array_res[,5,1,1]), pch=PCH.LIST[1], xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("number iterations")), ylim=c(0.5, 3.6)) #, ylim=range(log10(array_res[,5,,1])))
points(log10(array_res[,4,2,1]), log10(array_res[,5,2,1]), pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3,1]), log10(array_res[,5,3,1]), pch=PCH.LIST[3], col=3)
## points(log10(array_res[,4,1,2]), log10(array_res[,5,1,2]), pch=17, col=3)
## points(log10(array_res[,4,2,2]), log10(array_res[,5,2,2]), pch=18, col=4)
legend(0, 3.67, c("CG", "ULISSE q=0.1", "ULISSE q=1.0"), pch=PCH.LIST, col=c(1:3), cex=0.8)
dev.off()

## ************** VARIANCE vs CONDITION NUMBER
postscript(paste("PLOT_EARLYSTOP_GAMMA_", GAMMA,"_AVG_STDEV_vs_CONDITION_NUMBER.eps", sep=""))
par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
plot(log10(array_res[,4,1,1]), log10(array_res[,6,1,1]), type="n", ylim=range(log10(array_res[,6,2:3,1])), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("avg std dev")))
points(log10(array_res[,4,2,1]), log10(array_res[,6,2,1]), pch=PCH.LIST[2], col=2)
points(log10(array_res[,4,3,1]), log10(array_res[,6,3,1]), pch=PCH.LIST[3], col=3)
## points(log10(array_res[,4,1,2]), log10(array_res[,6,1,2]), pch=17, col=3)
## points(log10(array_res[,4,2,2]), log10(array_res[,6,2,2]), pch=18, col=4)
## legend(3.6, -5, c("CG", "ULISSE q=0.1", "ULISSE q=1.0")[2:3], pch=PCH.LIST[2:3], col=c(1:3)[2:3], cex=0.75)
dev.off()

## ## ************** N ITERATIONS vs CONDITION NUMBER
## postscript(paste("PLOT_EARLYSTOP_GAMMA_", GAMMA,"_NITERATIONS_vs_CONDITION_NUMBER.eps", sep=""))
## par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
## zz = ksmooth(log10(array_res[,4,1,1]), log10(array_res[,5,1,1]), kernel="normal", bandwidth=1)
## plot(zz$x, zz$y, type="l", ylim=range(log10(array_res[,5,,1])), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("number iterations")), lwd=12)
## points(log10(array_res[,4,1,1]), log10(array_res[,5,1,1]), pch=15, col=1)
## points(log10(array_res[,4,2,1]), log10(array_res[,5,2,1]), pch=16, col=2)
## points(log10(array_res[,4,3,1]), log10(array_res[,5,3,1]), pch=17, col=3)

## zz = ksmooth(log10(array_res[,4,2,1]), log10(array_res[,5,2,1]), kernel="normal", bandwidth=1)
## points(zz$x, zz$y, type="l", col=2, lwd=12)

## zz = ksmooth(log10(array_res[,4,3,1]), log10(array_res[,5,3,1]), kernel="normal", bandwidth=1)
## points(zz$x, zz$y, type="l", col=3, lwd=12)
## ## points(log10(array_res[,4,1,2]), log10(array_res[,5,1,2]), pch=17, col=3)
## ## points(log10(array_res[,4,2,2]), log10(array_res[,5,2,2]), pch=18, col=4)
## legend(0, 3.5, c("double", "q=0.1", "q=0.01"), pch=c(15:17), col=c(1:3), cex=0.8)
## dev.off()

## ## ## ************** VARIANCE vs CONDITION NUMBER
## postscript(paste("PLOT_EARLYSTOP_GAMMA_", GAMMA,"_AVG_STDEV_vs_CONDITION_NUMBER.eps", sep=""))
## par("mar"=c(3.0,3.0,1.1,0.3), "mgp"=c(1.8,0.6,0))
## zz = ksmooth(log10(array_res[,4,1,1]), log10(array_res[,6,1,1]), kernel="normal", bandwidth=1)
## plot(zz$x, zz$y, type="n", ylim=range(log10(array_res[,6,2:3,1])), xlab=expression(log[10](symbol(k))), ylab=expression(log[10]("avg std dev")), lwd=12)
## points(log10(array_res[,4,2,1]), log10(array_res[,6,2,1]), pch=16, col=2)
## points(log10(array_res[,4,3,1]), log10(array_res[,6,3,1]), pch=17, col=3)

## zz = ksmooth(log10(array_res[,4,2,1]), log10(array_res[,6,2,1]), kernel="normal", bandwidth=1)
## points(zz$x, zz$y, type="l", col=2, lwd=12)

## zz = ksmooth(log10(array_res[,4,3,1]), log10(array_res[,6,3,1]), kernel="normal", bandwidth=1)
## points(zz$x, zz$y, type="l", col=3, lwd=12)
## ## points(log10(array_res[,4,1,2]), log10(array_res[,6,1,2]), pch=17, col=3)
## ## points(log10(array_res[,4,2,2]), log10(array_res[,6,2,2]), pch=18, col=4)
## legend(4.0, -5, c("double", "q=0.1", "q=0.01")[2:3], pch=c(15:17)[2:3], col=c(1:3)[2:3], cex=0.8)
## dev.off()

}
