#define SPLIT_ROW_MATRIX 8

#define MAX_N_DIM 100

#define HOST_DESKTOP 0
#define HOST_CURIEUSE 1

#define PREC_FLOAT 0
#define PREC_DOUBLE 1

#define DEV_CPU 0
#define DEV_GPU 1

#define BY_DATA 0
#define BY_FEAT 1

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

#if (PRECISION == PREC_FLOAT)
  #if (DATA_ORIENT == BY_DATA)
    #define PROGRAM_FILE "../test_kernel_matrix_vector_multiplication/gpu_kernel_matrix_vector_multiplication_float_by_data.cl"
  #elif (DATA_ORIENT == BY_FEAT)
    #define PROGRAM_FILE "../test_kernel_matrix_vector_multiplication/gpu_kernel_matrix_vector_multiplication_float_by_feat.cl"
  #endif
#elif (PRECISION == PREC_DOUBLE)
  #if (DATA_ORIENT == BY_DATA)
    #define PROGRAM_FILE "../test_kernel_matrix_vector_multiplication/gpu_kernel_matrix_vector_multiplication_double_by_data.cl"
  #elif (DATA_ORIENT == BY_FEAT)
    #define PROGRAM_FILE "../test_kernel_matrix_vector_multiplication/gpu_kernel_matrix_vector_multiplication_double_by_feat.cl"
  #endif
#endif

#if (HOST==HOST_DESKTOP)
  #include <OpenCL/opencl.h>
#elif (HOST==HOST_CURIEUSE)
  #include <CL/opencl.h>
#endif


#include "../test_kernel_matrix_vector_multiplication/gpu_utilities.c"

#define MAX_LOOPS_CG 100000 /**< Maximum number of iterations for the conjugate gradient method. */


#if (PRECISION == PREC_FLOAT)

void kernel_matrix_vector_multiplication(int *n, cl_command_queue *command_queue, cl_kernel *kernel, cl_mem *v_gpu, float *v_float, cl_mem *res_gpu, float **res_float, float *exp_kpar_float, float *OMEGA_float, float *y) {

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
	clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n, v_float, 0, NULL, NULL);

  size_t global[2];
  global[0] = *n;
  global[1] = SPLIT_ROW_MATRIX;

  // enqueue the kernel command for execution
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueNDRangeKernel(command_queue[index_devices], kernel[index_devices], 2, NULL, &global[0], NULL, 0, NULL, NULL);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clFinish(command_queue[index_devices]);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n * SPLIT_ROW_MATRIX, res_float[index_devices], 0, NULL, NULL);

  int i, j;

  for(i = 0; i < *n; i++) {
    /* printf("%lf %lf \n",res_float[i], res_float[i + *n]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
		y[i] += res_float[index_devices][i + j * *n];
      }
    }
    y[i] += (exp_kpar_float[1] + *OMEGA_float) * v_float[i];
  }

}


#elif (PRECISION == PREC_DOUBLE)

void kernel_matrix_vector_multiplication(int *n, cl_command_queue *command_queue, cl_kernel *kernel, cl_mem *v_gpu, double *v_double, cl_mem *res_gpu, double **res_double, double *exp_kpar, double *OMEGA, double *y) {

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
	clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n, v_double, 0, NULL, NULL);

  size_t global[2];
  global[0] = *n;
  global[1] = SPLIT_ROW_MATRIX;

  // enqueue the kernel command for execution
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueNDRangeKernel(command_queue[index_devices], kernel[index_devices], 2, NULL, &global[0], NULL, 0, NULL, NULL);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clFinish(command_queue[index_devices]);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n * SPLIT_ROW_MATRIX, res_double[index_devices], 0, NULL, NULL);

  int i, j;

  for(i = 0; i < *n; i++) {
    /* printf("%lf %lf \n",res_double[i], res_double[i + *n]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
		y[i] += res_double[index_devices][i + j * *n];
      }
    }
    y[i] += (exp_kpar[1] + *OMEGA) * v_double[i];
  }

}

#endif



int solve_CG(double *X, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double *b_double, gsl_rng *uniform_random_variable, int n_results, double *result, double THRS_CONVERGENCE_CG, double THRS_EARLY_STOP_CG, double DECAY_EARLY_STOP_CG, int *niterations) {

  gsl_vector* b = gsl_vector_alloc(*n);
  for(int i = 0; i < *n; i++) gsl_vector_set(b, i, b_double[i]);

#if (PRECISION == PREC_FLOAT)
  
  float *X_float; X_float = (float*) malloc (*n * *d * sizeof(float));
  float *kpar_float; kpar_float = (float*) malloc (3 * sizeof(float));
  float *exp_kpar_float; exp_kpar_float = (float*) malloc (3 * sizeof(float));
  float *OMEGA_float; OMEGA_float = (float*) malloc (sizeof(float));
  /* float *v_float; v_float = (float*) malloc (*n * sizeof(float)); */

  float *res_float[N_DEVICES];  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    res_float[index_devices] = (float*) malloc (*n * SPLIT_ROW_MATRIX * sizeof(float));

#if (DATA_ORIENT == BY_DATA)
  for(int i = 0; i < *n; i++) {
    for(int j = 0; j < *d; j++) {
      X_float[i + j * *n] = (float) X[i + j * *n];
    }
  }
#elif (DATA_ORIENT == BY_FEAT)
  for(int i = 0; i < *d; i++) {
    for(int j = 0; j < *n; j++) {
      X_float[i + j * *d] = (float) X[i + j * *d];
    }
  }
#endif
  
  for(int i = 0; i < 3; i++) {
    kpar_float[i] = (float) kpar[i];
    exp_kpar_float[i] = (float) exp_kpar[i];
  }

  OMEGA_float[0] = (float) OMEGA[0];
  
#elif (PRECISION == PREC_DOUBLE)

  double *X_double; X_double = X;
  double *kpar_double; kpar_double = kpar;
  double *exp_kpar_double; exp_kpar_double = exp_kpar;
  double *OMEGA_double; OMEGA_double = OMEGA;
  /* double *v_double; */

  double *res_double[N_DEVICES];  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    res_double[index_devices] = (double*) malloc (*n * SPLIT_ROW_MATRIX * sizeof(double));

#endif

  cl_context context[N_DEVICES]; cl_program program[N_DEVICES]; cl_kernel kernel[N_DEVICES]; cl_command_queue command_queue[N_DEVICES]; 
  cl_uint num_of_platforms=0; cl_platform_id platform_id; cl_device_id device_id[N_DEVICES]; cl_uint num_of_devices=0; cl_int err; 

  cl_mem X_gpu[N_DEVICES], n_gpu[N_DEVICES], d_gpu[N_DEVICES], kpar_gpu[N_DEVICES], exp_kpar_gpu[N_DEVICES], OMEGA_gpu[N_DEVICES], v_gpu[N_DEVICES];
  cl_mem split_row_matrix_gpu[N_DEVICES], n_devices_gpu[N_DEVICES], gpu_id_gpu[N_DEVICES], res_gpu[N_DEVICES];

  int *split_row_matrix_pointer[N_DEVICES], *n_devices_pointer[N_DEVICES], *gpu_id_pointer[N_DEVICES];
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
    split_row_matrix_pointer[index_devices] = (int*) malloc (sizeof(int)); split_row_matrix_pointer[index_devices][0] = SPLIT_ROW_MATRIX;
    n_devices_pointer[index_devices] = (int*) malloc (sizeof(int)); n_devices_pointer[index_devices][0] = N_DEVICES;
    gpu_id_pointer[index_devices] = (int*) malloc (sizeof(int)); gpu_id_pointer[index_devices][0] = index_devices;
  }
  
  // retreive a list of platforms avaible
  if (clGetPlatformIDs(1, &platform_id, &num_of_platforms)!= CL_SUCCESS) { printf("Unable to get platform_id\n"); return 1; }
  
  // try to get a supported GPU device
#if (DEVICE == DEV_CPU)
  if (clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_CPU, N_DEVICES, device_id, &num_of_devices) != CL_SUCCESS) { printf("Unable to get device_id\n"); return 1; }
#elif (DEVICE == DEV_GPU)
  if (clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, N_DEVICES, device_id, &num_of_devices) != CL_SUCCESS) { printf("Unable to get device_id\n"); return 1; }
#endif

  /* printf("Detected %d devices - Using %d devices\n\n", num_of_devices, N_DEVICES); */
  
  // context properties list - must be terminated with 0
  cl_context_properties properties[3]; properties[0]= CL_CONTEXT_PLATFORM; properties[1]= (cl_context_properties) platform_id; properties[2]= 0;
  
  // create a context with the GPU device
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    context[index_devices] = clCreateContext(properties, N_DEVICES, device_id, NULL, NULL, &err);  

  // create command queue using the context and device
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    command_queue[index_devices] = clCreateCommandQueue(context[index_devices], device_id[index_devices], 0, &err);
  
  // create a program from the kernel source code
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    program[index_devices] = build_program(context[index_devices], device_id[index_devices], PROGRAM_FILE);

  // create buffers for the input and ouput
#if (PRECISION == PREC_FLOAT)
  // specify which kernel from the program to execute
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    kernel[index_devices] = clCreateKernel(program[index_devices], "matrix_vector_multiplication_float", &err);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    X_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * *n * *d, NULL, NULL);
    n_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    d_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * 3, NULL, NULL);
    exp_kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * 3, NULL, NULL);
    OMEGA_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float), NULL, NULL);
    v_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * *n, NULL, NULL);
    split_row_matrix_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    n_devices_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    gpu_id_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    res_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_WRITE_ONLY, sizeof(float) * *n * SPLIT_ROW_MATRIX, NULL, NULL);
  }

#elif (PRECISION == PREC_DOUBLE)
  // specify which kernel from the program to execute
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    kernel[index_devices] = clCreateKernel(program[index_devices], "matrix_vector_multiplication_double", &err);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    X_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * *n * *d, NULL, NULL);
    n_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    d_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * 3, NULL, NULL);
    exp_kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * 3, NULL, NULL);
    OMEGA_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double), NULL, NULL);
    v_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * *n, NULL, NULL);
    split_row_matrix_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    n_devices_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    gpu_id_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    res_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_WRITE_ONLY, sizeof(double) * *n * SPLIT_ROW_MATRIX, NULL, NULL);
  }

#endif
  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
  // set the argument list for the kernel command
    clSetKernelArg(kernel[index_devices], 0, sizeof(cl_mem), &X_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 1, sizeof(cl_mem), &n_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 2, sizeof(cl_mem), &d_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 3, sizeof(cl_mem), &kpar_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 4, sizeof(cl_mem), &exp_kpar_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 5, sizeof(cl_mem), &OMEGA_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 6, sizeof(cl_mem), &v_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 7, sizeof(cl_mem), &split_row_matrix_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 8, sizeof(cl_mem), &n_devices_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 9, sizeof(cl_mem), &gpu_id_gpu[index_devices]);
    clSetKernelArg(kernel[index_devices], 10, sizeof(cl_mem), &res_gpu[index_devices]);
  }


  // load data into the input buffer
#if (PRECISION == PREC_FLOAT)
  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    clEnqueueWriteBuffer(command_queue[index_devices], X_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n * *d, X_float, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], n_gpu[index_devices], CL_TRUE, 0, sizeof(int), n, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], d_gpu[index_devices], CL_TRUE, 0, sizeof(int), d, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], kpar_gpu[index_devices], CL_TRUE, 0, sizeof(float) * 3, kpar_float, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], exp_kpar_gpu[index_devices], CL_TRUE, 0, sizeof(float) * 3, exp_kpar_float, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], OMEGA_gpu[index_devices], CL_TRUE, 0, sizeof(float), OMEGA_float, 0, NULL, NULL);
    /* clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n, v_float, 0, NULL, NULL); */
    clEnqueueWriteBuffer(command_queue[index_devices], split_row_matrix_gpu[index_devices], CL_TRUE, 0, sizeof(int), split_row_matrix_pointer[index_devices], 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], n_devices_gpu[index_devices], CL_TRUE, 0, sizeof(int), n_devices_pointer[index_devices], 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], gpu_id_gpu[index_devices], CL_TRUE, 0, sizeof(int), gpu_id_pointer[index_devices], 0, NULL, NULL);
  }
  
#elif (PRECISION == PREC_DOUBLE)
  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    clEnqueueWriteBuffer(command_queue[index_devices], X_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n * *d, X_double, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], n_gpu[index_devices], CL_TRUE, 0, sizeof(int), n, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], d_gpu[index_devices], CL_TRUE, 0, sizeof(int), d, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, kpar_double, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], exp_kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, exp_kpar_double, 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], OMEGA_gpu[index_devices], CL_TRUE, 0, sizeof(double), OMEGA_double, 0, NULL, NULL);
    /* clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n, v_double, 0, NULL, NULL); */
    clEnqueueWriteBuffer(command_queue[index_devices], split_row_matrix_gpu[index_devices], CL_TRUE, 0, sizeof(int), split_row_matrix_pointer[index_devices], 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], n_devices_gpu[index_devices], CL_TRUE, 0, sizeof(int), n_devices_pointer[index_devices], 0, NULL, NULL);
    clEnqueueWriteBuffer(command_queue[index_devices], gpu_id_gpu[index_devices], CL_TRUE, 0, sizeof(int), gpu_id_pointer[index_devices], 0, NULL, NULL);
  }
  
#endif
  
  double u;
  
  int j;
  
  int n_after_early_stop[n_results];
  for(int i = 0; i < n_results; i++) {
    for(j = 0; j < MAX_LOOPS_CG; j++) {
      u = gsl_rng_uniform(uniform_random_variable);	  
      if(u > exp(-j / DECAY_EARLY_STOP_CG)) break;
    }
    n_after_early_stop[i] = j;

    /* printf("N EARLY STOP %d \n", n_after_early_stop[i]); */
  }

  double max_weighting_factor = 0, tmp_weighting_factor = 0;
  for(int i = 0; i < n_results; i++) {
	tmp_weighting_factor = exp( (n_after_early_stop[i] * (n_after_early_stop[i]-1)) / 2 / DECAY_EARLY_STOP_CG); // sum_{i=0}^(n-1) i = n * (n-1)/2
    if(tmp_weighting_factor > max_weighting_factor) max_weighting_factor = tmp_weighting_factor;
  }

  if(THRS_EARLY_STOP_CG == -999) THRS_EARLY_STOP_CG = sqrt(*n) / max_weighting_factor;
  
  int earlystop = 0, index_expansion = 0;
  
  // Initialize x using the vector in result (that is what is passed to the function as the value of K^-1 b from the previous iteration)
  gsl_vector* x = gsl_vector_alloc(*n);
  gsl_vector* result_gsl = gsl_vector_alloc(*n);
  for(int i = 0; i < *n; i++) gsl_vector_set(result_gsl, i, result[i]);
  gsl_vector_memcpy(x, result_gsl);
  
  
#if (PRECISION == PREC_FLOAT)
  
  float *x_float;
  x_float = (float*) malloc (*n * sizeof(float));
  for(int i = 0; i < *n; i++) x_float[i] = (float) gsl_vector_get(x, i);
  
#elif (PRECISION == PREC_DOUBLE)
  
  double *x_double;
  x_double = (double*) malloc (*n * sizeof(double));
  for(int i = 0; i < *n; i++) x_double[i] = gsl_vector_get(x, i);
  
#endif
  
  gsl_vector* x_unbiased = gsl_vector_alloc(*n);
  gsl_vector* p_tmp = gsl_vector_alloc(*n);
  double tmp;
  
  gsl_vector* Ax = gsl_vector_alloc(*n);
  
#if (PRECISION == PREC_FLOAT)
  float *Ax_float;
  Ax_float = (float*) malloc (*n * sizeof(float));
  
  kernel_matrix_vector_multiplication(n, command_queue, kernel, v_gpu, x_float, res_gpu, res_float, exp_kpar_float, OMEGA_float, Ax_float);
  for(int i = 0; i < *n; i++) gsl_vector_set(Ax, i, (double) Ax_float[i]);
  
#elif (PRECISION == PREC_DOUBLE)
  
  double *Ax_double;
  Ax_double = (double*) malloc (*n * sizeof(double));
  
  kernel_matrix_vector_multiplication(n, command_queue, kernel, v_gpu, x_double, res_gpu, res_double, exp_kpar, OMEGA, Ax_double);
  for(int i = 0; i < *n; i++) gsl_vector_set(Ax, i, Ax_double[i]);
  
#endif


  /* // r = b - Ax */
  gsl_vector* r = gsl_vector_alloc(*n);
  gsl_vector_memcpy(r, b);
  gsl_vector_sub(r, Ax);

  gsl_vector* q = gsl_vector_alloc(*n);

  // p = b - Ax
  gsl_vector* p = gsl_vector_alloc(*n);
  gsl_vector_memcpy(p, r);

#if (PRECISION == PREC_FLOAT)
  float *p_float;
  p_float = (float*) malloc (*n * sizeof(float));
  float *Ap_float;
  Ap_float = (float*) malloc (*n * sizeof(float));
#elif (PRECISION == PREC_DOUBLE)
  double *p_double;
  p_double = (double*) malloc (*n * sizeof(double));
  double *Ap_double;
  Ap_double = (double*) malloc (*n * sizeof(double));
#endif

  double diff;
  double alphaN = 0;
  double alphaD = 0;
  double alpha = 0;
  double betaN = 0;
  double beta = 0;
  gsl_vector* Ap = gsl_vector_alloc(*n);
  
  int k = 0;
  do {
    // alpha = rTr / pAp
    gsl_blas_ddot(r, r, &alphaN);
	
#if (PRECISION == PREC_FLOAT)
    for(int i = 0; i < *n; i++) p_float[i] = (float) gsl_vector_get(p, i);
    kernel_matrix_vector_multiplication(n, command_queue, kernel, v_gpu, (float *) p_float, res_gpu, res_float, exp_kpar_float, OMEGA_float, Ap_float);
    for(int i = 0; i < *n; i++) gsl_vector_set(Ap, i, (double) Ap_float[i]);
#elif (PRECISION == PREC_DOUBLE)
    for(int i = 0; i < *n; i++) p_double[i] = gsl_vector_get(p, i);
    kernel_matrix_vector_multiplication(n, command_queue, kernel, v_gpu, p_double, res_gpu, res_double, exp_kpar, OMEGA, Ap_double);
    for(int i = 0; i < *n; i++) gsl_vector_set(Ap, i, Ap_double[i]);
#endif
    
    gsl_blas_ddot(p, Ap, &alphaD);
    alpha = alphaN / alphaD;
    
    gsl_vector_memcpy(q, r);
    
    // x = x + alpha*p
    gsl_vector_scale(p, alpha);
    gsl_vector_add(x, p);
    
    // r = r - alpha*Ap
    gsl_vector_scale(Ap, alpha);
    gsl_vector_sub(r, Ap);
    
    gsl_vector_sub(q, r);
    gsl_blas_ddot(q, q, &diff);
    
    if(sqrt(diff) < THRS_CONVERGENCE_CG) {
      /* gsl_vector_memcpy(result_gsl, x); */
      for(unsigned int i = 0; i < n_results; i++) {
		for(unsigned int j = 0; j < *n; j++)
		  result[j + i * *n] = gsl_vector_get(x, j);
      }
      break;
    }
    
    if(earlystop == 0) {
      if(sqrt(diff) < THRS_EARLY_STOP_CG) {
		earlystop = 1;
		gsl_vector_memcpy(x_unbiased, x);
		gsl_vector_sub(x_unbiased, p);
		index_expansion = 0;
      }
    }
    
    if(earlystop == 1) {
      int checkpoint = 1;
      for(unsigned int i = 0; i < n_results; i++) checkpoint *= (index_expansion >= n_after_early_stop[i]);
      if(checkpoint == 1) {
		break;
      }
      
      else {
		gsl_vector_memcpy(p_tmp, p);
		/* tmp = 1.0; */ // Explicit (old fashioned) way to compute the weighting factor for the terms in the expansion
		/* for(unsigned int i = 0; i < index_expansion; i++) tmp *= exp(i / DECAY_EARLY_STOP_CG); */
		
		/* printf("%d %lf %lf\n", index_expansion, tmp, exp(index_expansion * (index_expansion-1) / 2 / DECAY_EARLY_STOP_CG) ); */

		tmp = exp(index_expansion * (index_expansion-1) / 2 / DECAY_EARLY_STOP_CG); // Weighting factor for the terms in the expansion - prod_i exp(i / DECAY_EARLY_STOP_CG) = exp((sum_i i) / DECAY_EARLY_STOP_CG)

		gsl_vector_scale(p_tmp, tmp);
		gsl_vector_add(x_unbiased, p_tmp);
		/* gsl_vector_memcpy(result_gsl, x_unbiased); */
		
		for(unsigned int i = 0; i < n_results; i++) {
		  if( index_expansion == (n_after_early_stop[i]-1) ) {
			for(unsigned int j = 0; j < *n; j++) {
			  result[j + i * *n] = gsl_vector_get(x_unbiased, j);
			  niterations[i] = k + 2;
			}
		  }
		}
      }
      
      index_expansion++;
    }
    
    // beta = (r1Tr1 / r0Tr0) * alpha^(-1)
    gsl_blas_ddot(r, r, &betaN);
    beta = (betaN * alphaD) / (alphaN * alphaN);
    
    // p = beta*p + r
    gsl_vector_scale(p, beta);
    gsl_vector_add(p, r);
    k++;
    
  } while(k < MAX_LOOPS_CG);

  if(k == MAX_LOOPS_CG) {
    gsl_vector_memcpy(result_gsl, x);
    printf("CG reached limit of %d \n", MAX_LOOPS_CG);
    for(int i = 0; i < *n; i++) result[i] = gsl_vector_get(result_gsl, i);
  }
  
  
  /* printf("\n\nFinished in %d iterations\n\n", k); */

  for(unsigned int i = 0; i < n_results; i++) { 
	if(niterations[i] == 0) niterations[i] = k + 2;
  }

  printf("%d ", k + 2);
  
  gsl_vector_free(r);
  gsl_vector_free(x);
  gsl_vector_free(x_unbiased);
  gsl_vector_free(p);
  gsl_vector_free(p_tmp);
  gsl_vector_free(Ap);

  gsl_vector_free(Ax);
  gsl_vector_free(result_gsl);


#if (PRECISION == PREC_FLOAT)

  free(X_float);
  free(kpar_float);
  free(exp_kpar_float);
  free(OMEGA_float);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    free(res_float[index_devices]);
  }

  free(x_float);
  free(p_float);
  free(Ax_float);
  free(Ap_float);

#elif (PRECISION == PREC_DOUBLE)

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    free(res_double[index_devices]);
  }

  free(x_double);
  free(p_double);
  free(Ax_double);
  free(Ap_double);

#endif

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    clReleaseMemObject(X_gpu[index_devices]);
 
    clReleaseMemObject(d_gpu[index_devices]);
    clReleaseMemObject(n_gpu[index_devices]);
    clReleaseMemObject(kpar_gpu[index_devices]);
    clReleaseMemObject(exp_kpar_gpu[index_devices]);
    clReleaseMemObject(OMEGA_gpu[index_devices]);
    clReleaseMemObject(v_gpu[index_devices]);
    clReleaseMemObject(split_row_matrix_gpu[index_devices]);
    clReleaseMemObject(n_devices_gpu[index_devices]);
    clReleaseMemObject(gpu_id_gpu[index_devices]);
    clReleaseMemObject(res_gpu[index_devices]);

    clReleaseProgram(program[index_devices]);
    clReleaseKernel(kernel[index_devices]);
    clReleaseCommandQueue(command_queue[index_devices]);
    clReleaseContext(context[index_devices]);
  }

  return 0;
}


int wrapper_linear_system(double *DATA, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double *b, int *n_results, double *result, double *THRS_CONVERGENCE_CG_pointer, double *THRS_EARLY_STOP_CG_pointer, double *DECAY_EARLY_STOP_CG_pointer, int *niterations) {

  if(*d > MAX_N_DIM) { printf("PROBLEM DIMENSION IS TOO LARGE - CHANGE THE FUNCTION THAT IS EXECUTED ON THE GPU TO ALLOW FOR LARGER DIMENSIONS \n\n\n"); return 1; }
  
/* #if (HOST==HOST_DESKTOP) */
/*   printf("---- COMPILED FOR DESKTOP ----"); */
/* #elif (HOST==HOST_CURIEUSE) */
/*   printf("---- COMPILED FOR CURIEUSE ----"); */
/* #endif */
  
/* #if (PRECISION == PREC_FLOAT) */
/*   printf("---- FLOAT ----"); */
/* #elif (PRECISION == PREC_DOUBLE) */
/*   printf("---- DOUBLE ----"); */
/* #endif */
  
/* #if (DEVICE == DEV_CPU) */
/*   printf("---- CL CPU ----"); */
/* #elif (DEVICE == DEV_GPU) */
/*   printf("---- CL GPU ----"); */
/* #endif */
  
/* #if (DATA_ORIENT == BY_DATA) */
/*   printf("---- BY DATA ----"); */
/* #elif (DATA_ORIENT == BY_FEAT) */
/*   printf("---- BY FEAT ----"); */
/* #endif */

  int i, j;

  // Initialize random number generator
  const gsl_rng_type *type_random_sampling;
  gsl_rng *uniform_random_variable;

  type_random_sampling = gsl_rng_mt19937;
  uniform_random_variable = gsl_rng_alloc (type_random_sampling);
  gsl_rng_set(uniform_random_variable, time (NULL));
  // **********************************

  solve_CG(DATA, n, d, kpar, exp_kpar, OMEGA, b, uniform_random_variable, *n_results, result, *THRS_CONVERGENCE_CG_pointer, *THRS_EARLY_STOP_CG_pointer, *DECAY_EARLY_STOP_CG_pointer, niterations);

  return 0;
}
