SEED = 1

SHOW.ACCURACY = F

MODE = "cl_cpu_by_feat"
## MODE = "cl_gpu_by_feat"

PRECISION = "FLOAT"
PRECISION = "DOUBLE"

cat("MODE ", MODE, "  ")
cat("PRECISION ", PRECISION, "  ")

if(MODE == "cl_cpu_by_feat") {
  if(PRECISION == "FLOAT") dyn.load("sgr_cpu_float_by_feat.so")
  if(PRECISION == "DOUBLE") dyn.load("sgr_cpu_double_by_feat.so")
}

if(MODE == "cl_gpu_by_feat") {
  if(PRECISION == "FLOAT") dyn.load("sgr_gpu_float_by_feat.so")
  if(PRECISION == "DOUBLE") dyn.load("sgr_gpu_double_by_feat.so")
}

PRIOR.SIGMA.SHAPE = 1
PRIOR.SIGMA.RATE = 0.1
PRIOR.LAMBDA.SHAPE = 1
PRIOR.LAMBDA.RATE = 0.1
PRIOR.TAU.SHAPE = 1
PRIOR.TAU.RATE = 0.1
PRIORS = c(PRIOR.SIGMA.SHAPE, PRIOR.SIGMA.RATE, PRIOR.LAMBDA.SHAPE, PRIOR.LAMBDA.RATE, PRIOR.TAU.SHAPE, PRIOR.TAU.RATE)

OMEGA = 1e-9

NPAR = 3

NRVECT = 4

## ******************** DATASET
DATA = new.env()
DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_X.txt")))
DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/concrete_y.txt"))

## DATA$X = t(as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_X.txt"))[1:1000,])
## DATA$y = as.matrix(read.table("~/DATASET_REGRESSION/clean/census8L_y.txt"))[1:1000]

DATA$n = dim(DATA$X)[2]
DATA$d = dim(DATA$X)[1]

THRS_CONVERGENCE_CG = 1.0e-8
THRS_EARLY_STOP_CG = sqrt(DATA$n) * 1
DECAY_EARLY_STOP_CG = 1.0


## ******************** SUBSET DATA for MAP ESTIMATION
cat("MAP on subset of data... ")

loglik = function(theta) {
  K = exp( - exp(theta[3]) * as.matrix(dist(SUBSET.DATA$X)^2) + theta[1]) + diag(exp(theta[2]) + 1e-9, SUBSET.DATA$n)
  L.chol = t(chol(K))
  K.inv.y = backsolve(t(L.chol), forwardsolve(L.chol, SUBSET.DATA$y))

  - sum(log(diag(L.chol))) - 0.5 * crossprod(SUBSET.DATA$y, K.inv.y) + PRIOR.SIGMA.SHAPE * theta[1] - PRIOR.SIGMA.RATE * exp(theta[1]) + PRIOR.LAMBDA.SHAPE * theta[2] - PRIOR.LAMBDA.RATE * exp(theta[2]) + sum(PRIOR.TAU.SHAPE * theta[3] - PRIOR.TAU.RATE * exp(theta[3]))
}

set.seed(21323)
ind.subset = sample(c(1:DATA$n), 500, replace=F)
SUBSET.DATA = new.env()
SUBSET.DATA$X = t(DATA$X)[ind.subset,]
SUBSET.DATA$y = DATA$y[ind.subset]
SUBSET.DATA$n = dim(SUBSET.DATA$X)[1]
SUBSET.DATA$d = dim(SUBSET.DATA$X)[2]

best.model = list(value=-Inf)
for(i in 1:5) {
  cat(i, " ")
  model = optim(runif(NPAR, min=c(0,-2, -5), max=c(5, 2, 3)), loglik, method="L-BFGS", control=list(maxit=1000, fnscale=-1))
  if(model$value > best.model$value) best.model = model
}

hessian = matrix(0, NPAR, NPAR)
deltax = 1e-4
for(i in 1:NPAR) {

 evect = rep(0, NPAR)
 evect[i] = 1

 hessian[i, i] = (loglik(best.model$par + deltax * evect) + loglik(best.model$par - deltax * evect) - 2 * loglik(best.model$par)) / deltax^2
}

for(i in 1:(NPAR-1)) {
   ei = rep(0, NPAR); ei[i] = 1
   for(j in (i+1):NPAR) {
       ej = rep(0, NPAR); ej[j] = 1

       hessian[i, j] = hessian[j, i] = (loglik(best.model$par + deltax * (ei + ej) ) - loglik(best.model$par - deltax * (ei - ej)) + loglik(best.model$par + deltax * (-ei - ej) ) - loglik(best.model$par - deltax * (-ei + ej)) ) / (4 * deltax^2)
   }
}

M_MAT = -solve(hessian)
tmp = eigen(M_MAT)
M12 = tmp[[2]] %*% diag(sqrt(tmp[[1]])) %*% t(tmp[[2]])



cat("done\n\n")
cat("MAP parameters on subset of data", best.model$par, "\n\n")

MAXIT = 40000
GAMMA = 1

EMAX = 0.1
EMIN = EMAX / 1e3

set.seed(184429 + 77 * SEED)

INIT_PAR = M12 %*% runif(3) + best.model$par ## best.model$par
EXP_INIT_PAR = exp(INIT_PAR)

SAMPLES = matrix(0, MAXIT+1, NPAR)
SAMPLES_GRAD = matrix(0, MAXIT+1, NPAR)
SAMPLES_DIAG_G = matrix(0, MAXIT+1, NPAR)
SAMPLES[1,] = INIT_PAR

gradient_final = rep(0, NPAR)
rvect = matrix(0, DATA$n, NRVECT)

running.time = system.time(assign("res", .C("sgr", as.double(DATA$X), as.double(DATA$y), as.integer(DATA$n), as.integer(DATA$d), as.double(INIT_PAR), as.double(EXP_INIT_PAR), as.integer(NPAR), as.double(OMEGA), as.integer(MAXIT), as.double(GAMMA), as.double(EMIN), as.double(EMAX), as.double(PRIORS), as.double(gradient_final), as.double(rvect), as.double(SAMPLES), as.double(SAMPLES_GRAD), as.double(M12), as.integer(NRVECT), as.double(THRS_CONVERGENCE_CG), as.double(THRS_EARLY_STOP_CG), as.double(DECAY_EARLY_STOP_CG), as.integer(SEED))))

## SS$psi.sigma = res[[5]][1]
## SS$psi.lambda = res[[5]][2]
## SS$psi.tau = res[[5]][3]
## compute.gradient(SS)
SAMPLES = matrix(res[[16]], ncol = NPAR)
SAMPLES_GRAD = matrix(res[[17]], ncol = NPAR)


save.image(paste("RES_REAL_DATA_SMALL_", SEED, ".Rdata", sep=""))

## par(mfrow=c(1,3))

## BBB = MAXIT / ((EMAX/EMIN)^(1 / GAMMA) - 1)
## AAA = EMAX * BBB^(GAMMA)
## epsilon.plot = AAA / (BBB + c(0:MAXIT))^(GAMMA)

## plot(cumsum(SAMPLES[,1]*epsilon.plot)/cumsum(epsilon.plot), type="l", main="posterior expectation of psi.sigma")
## ## abline(h = TRUE.PSI.SIGMA)

## plot(cumsum(SAMPLES[,2]*epsilon.plot)/cumsum(epsilon.plot), type="l", main="posterior expectation of psi.lambda")
## ## abline(h = TRUE.PSI.LAMBDA)

## plot(cumsum(SAMPLES[,3]*epsilon.plot)/cumsum(epsilon.plot), type="l", main="posterior expectation of psi.tau")
## ## abline(h = TRUE.PSI.TAU)



if(SHOW.ACCURACY == T) {
  gradient_final = res[[14]]
  rvect = res[[15]]
  theta = SAMPLES[MAXIT+1,]
  exp_theta = exp(theta)
  
  X = DATA$X
  y = DATA$y
  n = DATA$n
  
  K = exp(theta[1]) * exp( - exp(theta[3]) * as.matrix(dist(t(X))^2)) + diag(exp(theta[2]) + OMEGA, n)
  
  dKdt1 = exp(theta[1]) * exp( - exp(theta[3]) * as.matrix(dist(t(X))^2))
  dKdt2 = diag(exp_theta[2], n)
  dKdt3 = dKdt1 * (log(dKdt1) - theta[1])
  
  
  grad_R = rep(0, 3)
  
  grad_R[1] = - 0.5 * t(rvect) %*% solve(K) %*% dKdt1 %*% rvect + 0.5 * t(y) %*% solve(K) %*% dKdt1 %*% solve(K) %*% y + PRIOR.SIGMA.SHAPE - PRIOR.SIGMA.RATE * exp_theta[1];
  grad_R[2] = - 0.5 * t(rvect) %*% solve(K) %*% dKdt2 %*% rvect + 0.5 * t(y) %*% solve(K) %*% dKdt2 %*% solve(K) %*% y + PRIOR.LAMBDA.SHAPE - PRIOR.LAMBDA.RATE * exp_theta[2];
  grad_R[3] = - 0.5 * t(rvect) %*% solve(K) %*% dKdt3 %*% rvect + 0.5 * t(y) %*% solve(K) %*% dKdt3 %*% solve(K) %*% y + PRIOR.TAU.SHAPE - PRIOR.TAU.RATE * exp_theta[3];
}



break

## nrepetitions = 2

## acceptance.rate.plot = rep(0, MAXIT+1)
## previous.log.joint = SS$log.p.y.giv.theta + SS$log.p.theta


## epsilon.plot = rep(0, MAXIT+1)
## var.stochastic.gradients.plot = matrix(0, NPAR, MAXIT+1)

## for(ttt in 0:MAXIT) {

##   cat(ttt, "    \r")
  
##   epsilon = AAA / (BBB + ttt)^(GAMMA)
##   epsilon.plot[ttt+1] = epsilon

##   stochastic.gradients = matrix(0, nrepetitions, NPAR)
##   for(repetition in 1:nrepetitions) {
##     compute.gradient(SS)
##     stochastic.gradients[repetition,] = SS$GRAD.LOGJOINTLIK1 
##   }
##   var.stochastic.gradients = diag(cov(epsilon / 2 * stochastic.gradients))
##   var.stochastic.gradients.plot[,ttt+1] = var.stochastic.gradients

##   acceptance.rates = rep(0, nrepetitions)
##   for(repetition in 1:nrepetitions) {
##     newpar = epsilon / 2 * stochastic.gradients[repetition,] + rnorm(NPAR, sd = sqrt(epsilon))

##     SS.TMP$psi.sigma = SS$psi.sigma + newpar[1]
##     SS.TMP$psi.lambda = SS$psi.lambda + newpar[2]
##     SS.TMP$psi.tau = SS$psi.tau + newpar[3]

##     compute.gradient(SS.TMP)

##     current.log.joint = SS.TMP$log.p.y.giv.theta + SS.TMP$log.p.theta
##     acceptance.rates[repetition] = min(exp(current.log.joint - previous.log.joint), 1)
##   }
  
##   acceptance.rate.plot[ttt + 1] = mean(acceptance.rates)
  
##   SS$psi.sigma = SS.TMP$psi.sigma
##   SS$psi.lambda = SS.TMP$psi.lambda
##   SS$psi.tau = SS.TMP$psi.tau
##   previous.log.joint = current.log.joint

##   SAMPLES[ttt+1,] = c(SS.TMP$psi.sigma, SS.TMP$psi.lambda, SS.TMP$psi.tau)
  
##   ## print(c(SS$psi.sigma, SS$psi.lambda, SS$psi.tau))
## }

## par(mfrow=c(2,3))
## plot(log10(0:MAXIT), log10(epsilon.plot), type="l", col=2, ylim=range(log10(c(var.stochastic.gradients.plot, epsilon.plot))))
## points(log10(0:MAXIT), log10(var.stochastic.gradients.plot[1,]), type="l", col=3)
## points(log10(0:MAXIT), log10(var.stochastic.gradients.plot[2,]), type="l", col=4)
## points(log10(0:MAXIT), log10(var.stochastic.gradients.plot[3,]), type="l", col=5)

## plot(log10(0:MAXIT), log10(1 - acceptance.rate.plot), type="l", ylab = "log10 rejection rate")

## plot(c(1))

## plot(cumsum(SAMPLES[,1]*epsilon.plot)/cumsum(epsilon.plot), type="l", main="posterior expectation of psi.sigma")
## plot(cumsum(SAMPLES[,2]*epsilon.plot)/cumsum(epsilon.plot), type="l", main="posterior expectation of psi.lambda")
## plot(cumsum(SAMPLES[,3]*epsilon.plot)/cumsum(epsilon.plot), type="l", main="posterior expectation of psi.tau")

