#define OUTPUT_FILE "SAMPLES"
#define OUTPUT_FILE_GRAD "SAMPLES_GRAD"
#define OUTPUT_FILE_RATIO_VAR "SAMPLES_RATIO_VAR"
#define OUTPUT_FILE_EPSILON "SAMPLES_EPSILON"


#define SPLIT_ROW_MATRIX 8

#define MAX_N_DIM 100

#define HOST_DESKTOP 0
#define HOST_CURIEUSE 1

#define PREC_FLOAT 0
#define PREC_DOUBLE 1

#define DEV_CPU 0
#define DEV_GPU 1

#define BY_DATA 0
#define BY_FEAT 1

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>
#if (HOST==HOST_DESKTOP)
  #include <OpenCL/opencl.h>
#elif (HOST==HOST_CURIEUSE)
  #include <CL/opencl.h>
#endif


#if (PRECISION == PREC_FLOAT)
  #if (DATA_ORIENT == BY_DATA)
    #define PROGRAM_FILE "gpu_kernels_float_by_data.cl"
  #elif (DATA_ORIENT == BY_FEAT)
    #define PROGRAM_FILE "gpu_kernels_float_by_feat.cl"
  #endif
#elif (PRECISION == PREC_DOUBLE)
  #if (DATA_ORIENT == BY_DATA)
    #define PROGRAM_FILE "gpu_kernels_double_by_data.cl"
  #elif (DATA_ORIENT == BY_FEAT)
    #define PROGRAM_FILE "gpu_kernels_double_by_feat.cl"
  #endif
#endif


#include "../test_kernel_matrix_vector_multiplication/gpu_utilities.c"

#define MAX_LOOPS_CG 10000 /**< Maximum number of iterations for the conjugate gradient method. */

#include "utility_dK_vector_multiplication.c"
#include "utility_linear_system.c"
#include "utility_gradient.c"







int sgr(double *X, double *y, int *n, int *d, double *kpar, double *exp_kpar, int *npar, double *OMEGA, int *MAXIT, double *GAMMA, double *EPSILON_MIN, double *EPSILON_MAX, double *priors, double *gradient_final, double *rvect, double *SAMPLES, double *SAMPLES_GRAD, double *M12, int *NRVECT, double *THRS_CONVERGENCE_CG_pointer, double *THRS_EARLY_STOP_CG_pointer, double *DECAY_EARLY_STOP_CG_pointer, int *SEED) {

  int i, j;

#if (PRECISION == PREC_FLOAT)
  
  float *X_float; X_float = (float*) malloc (*n * *d * sizeof(float));
  float *kpar_float; kpar_float = (float*) malloc (3 * sizeof(float));
  float *exp_kpar_float; exp_kpar_float = (float*) malloc (3 * sizeof(float));
  float *OMEGA_float; OMEGA_float = (float*) malloc (sizeof(float));
  float *v_float; v_float = (float*) malloc (*n * sizeof(float));

  float *res_float[N_DEVICES];  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    res_float[index_devices] = (float*) malloc (*n * SPLIT_ROW_MATRIX * sizeof(float));

#if (DATA_ORIENT == BY_DATA)
  for(i = 0; i < *n; i++) {
    for(j = 0; j < *d; j++) {
      X_float[i + j * *n] = (float) X[i + j * *n];
    }
  }
#elif (DATA_ORIENT == BY_FEAT)
  for(i = 0; i < *d; i++) {
    for(j = 0; j < *n; j++) {
      X_float[i + j * *d] = (float) X[i + j * *d];
    }
  }
#endif
  
  for(i = 0; i < 3; i++) {
    kpar_float[i] = (float) kpar[i];
    exp_kpar_float[i] = (float) exp_kpar[i];
  }

  OMEGA_float[0] = (float) OMEGA[0];
  
  for(i = 0; i < *n; i++) {
    v_float[i] = (float) v[i];
  }

#elif (PRECISION == PREC_DOUBLE)

  double *X_double; X_double = X;
  double *kpar_double; kpar_double = kpar;
  double *exp_kpar_double; exp_kpar_double = exp_kpar;
  double *OMEGA_double; OMEGA_double = OMEGA;
  /* double *v_double; // v_double = v; */

  double *kpar_tmp; kpar_tmp = (double*) malloc (*n * sizeof(double));
  int norm_kpar_flag = 0;

  double *res_double[N_DEVICES];  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    res_double[index_devices] = (double*) malloc (*n * SPLIT_ROW_MATRIX * sizeof(double));

#endif

  cl_context context[N_DEVICES]; 
  cl_program program[N_DEVICES];
  cl_kernel kernel_kernel_matrix_vector_multiplication[N_DEVICES]; cl_kernel kernel_dK_vector_multiplication_psi_sigma[N_DEVICES]; cl_kernel kernel_dK_vector_multiplication_psi_tau[N_DEVICES]; 
  cl_command_queue command_queue[N_DEVICES]; 
  cl_uint num_of_platforms=0; cl_platform_id platform_id; cl_device_id device_id[N_DEVICES]; cl_uint num_of_devices=0; cl_int err; 

  cl_mem X_gpu[N_DEVICES], n_gpu[N_DEVICES], d_gpu[N_DEVICES], kpar_gpu[N_DEVICES], exp_kpar_gpu[N_DEVICES], OMEGA_gpu[N_DEVICES], v_gpu[N_DEVICES];
  cl_mem split_row_matrix_gpu[N_DEVICES], n_devices_gpu[N_DEVICES], gpu_id_gpu[N_DEVICES], res_gpu[N_DEVICES];

  int *split_row_matrix_pointer[N_DEVICES], *n_devices_pointer[N_DEVICES], *gpu_id_pointer[N_DEVICES];
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
    split_row_matrix_pointer[index_devices] = (int*) malloc (sizeof(int)); split_row_matrix_pointer[index_devices][0] = SPLIT_ROW_MATRIX;
    n_devices_pointer[index_devices] = (int*) malloc (sizeof(int)); n_devices_pointer[index_devices][0] = N_DEVICES;
    gpu_id_pointer[index_devices] = (int*) malloc (sizeof(int)); gpu_id_pointer[index_devices][0] = index_devices;
  }
  
  // retreive a list of platforms avaible
  if (clGetPlatformIDs(1, &platform_id, &num_of_platforms)!= CL_SUCCESS) { printf("Unable to get platform_id\n"); return 1; }
  
  // try to get a supported GPU device
#if (DEVICE == DEV_CPU)
  if (clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_CPU, N_DEVICES, device_id, &num_of_devices) != CL_SUCCESS) { printf("Unable to get device_id\n"); return 1; }
#elif (DEVICE == DEV_GPU)
  if (clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, N_DEVICES, device_id, &num_of_devices) != CL_SUCCESS) { printf("Unable to get device_id\n"); return 1; }
#endif

  // context properties list - must be terminated with 0
  cl_context_properties properties[3]; properties[0]= CL_CONTEXT_PLATFORM; properties[1]= (cl_context_properties) platform_id; properties[2]= 0;
  
  // create a context with the GPU device
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    context[index_devices] = clCreateContext(properties, N_DEVICES, device_id, NULL, NULL, &err);  

  // create command queue using the context and device
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) 
    command_queue[index_devices] = clCreateCommandQueue(context[index_devices], device_id[index_devices], 0, &err);
  
  // create a program from the kernel source code
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
    program[index_devices] = build_program(context[index_devices], device_id[index_devices], PROGRAM_FILE);
  }

  // create buffers for the input and ouput
#if (PRECISION == PREC_FLOAT)
  // specify which kernel from the program to execute
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
    kernel_kernel_matrix_vector_multiplication[index_devices] = clCreateKernel(program[index_devices], "matrix_vector_multiplication_float", &err);

    kernel_dK_vector_multiplication_psi_sigma[index_devices] = clCreateKernel(program[index_devices], "dK_vector_multiplication_float_psi_sigma", &err);
    kernel_dK_vector_multiplication_psi_tau[index_devices] = clCreateKernel(program[index_devices], "dK_vector_multiplication_float_psi_tau", &err);
  }
  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    X_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * *n * *d, NULL, NULL);
    n_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    d_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * 3, NULL, NULL);
    exp_kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * 3, NULL, NULL);
    OMEGA_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float), NULL, NULL);
    v_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(float) * *n, NULL, NULL);
    split_row_matrix_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    n_devices_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    gpu_id_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    res_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_WRITE_ONLY, sizeof(float) * *n * SPLIT_ROW_MATRIX, NULL, NULL);
  }
  
#elif (PRECISION == PREC_DOUBLE)
  // specify which kernel from the program to execute
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
    kernel_kernel_matrix_vector_multiplication[index_devices] = clCreateKernel(program[index_devices], "matrix_vector_multiplication_double", &err);

    kernel_dK_vector_multiplication_psi_sigma[index_devices] = clCreateKernel(program[index_devices], "dK_vector_multiplication_double_psi_sigma", &err);
	kernel_dK_vector_multiplication_psi_tau[index_devices] = clCreateKernel(program[index_devices], "dK_vector_multiplication_double_psi_tau", &err);
  }
  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    X_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * *n * *d, NULL, NULL);
    n_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    d_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * 3, NULL, NULL);
    exp_kpar_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * 3, NULL, NULL);
    OMEGA_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double), NULL, NULL);
    v_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(double) * *n, NULL, NULL);
    split_row_matrix_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    n_devices_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    gpu_id_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_READ_ONLY, sizeof(int), NULL, NULL);
    res_gpu[index_devices] = clCreateBuffer(context[index_devices], CL_MEM_WRITE_ONLY, sizeof(double) * *n * SPLIT_ROW_MATRIX, NULL, NULL);
  }

#endif
  
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
  // set the argument list for the kernel command
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 0, sizeof(cl_mem), &X_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 1, sizeof(cl_mem), &n_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 2, sizeof(cl_mem), &d_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 3, sizeof(cl_mem), &kpar_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 4, sizeof(cl_mem), &exp_kpar_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 5, sizeof(cl_mem), &OMEGA_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 6, sizeof(cl_mem), &v_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 7, sizeof(cl_mem), &split_row_matrix_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 8, sizeof(cl_mem), &n_devices_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 9, sizeof(cl_mem), &gpu_id_gpu[index_devices]);
    clSetKernelArg(kernel_kernel_matrix_vector_multiplication[index_devices], 10, sizeof(cl_mem), &res_gpu[index_devices]);

    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 0, sizeof(cl_mem), &X_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 1, sizeof(cl_mem), &n_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 2, sizeof(cl_mem), &d_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 3, sizeof(cl_mem), &kpar_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 4, sizeof(cl_mem), &exp_kpar_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 5, sizeof(cl_mem), &OMEGA_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 6, sizeof(cl_mem), &v_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 7, sizeof(cl_mem), &split_row_matrix_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 8, sizeof(cl_mem), &n_devices_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 9, sizeof(cl_mem), &gpu_id_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_sigma[index_devices], 10, sizeof(cl_mem), &res_gpu[index_devices]);

    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 0, sizeof(cl_mem), &X_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 1, sizeof(cl_mem), &n_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 2, sizeof(cl_mem), &d_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 3, sizeof(cl_mem), &kpar_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 4, sizeof(cl_mem), &exp_kpar_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 5, sizeof(cl_mem), &OMEGA_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 6, sizeof(cl_mem), &v_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 7, sizeof(cl_mem), &split_row_matrix_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 8, sizeof(cl_mem), &n_devices_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 9, sizeof(cl_mem), &gpu_id_gpu[index_devices]);
    clSetKernelArg(kernel_dK_vector_multiplication_psi_tau[index_devices], 10, sizeof(cl_mem), &res_gpu[index_devices]);
  }


  // load data into the input buffer
#if (PRECISION == PREC_FLOAT)

    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      clEnqueueWriteBuffer(command_queue[index_devices], X_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n * *d, X_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_gpu[index_devices], CL_TRUE, 0, sizeof(int), n, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], d_gpu[index_devices], CL_TRUE, 0, sizeof(int), d, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], kpar_gpu[index_devices], CL_TRUE, 0, sizeof(float) * 3, kpar_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], exp_kpar_gpu[index_devices], CL_TRUE, 0, sizeof(float) * 3, exp_kpar_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], OMEGA_gpu[index_devices], CL_TRUE, 0, sizeof(float), OMEGA_float, 0, NULL, NULL);
      //      clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n, v_float, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], split_row_matrix_gpu[index_devices], CL_TRUE, 0, sizeof(int), split_row_matrix_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_devices_gpu[index_devices], CL_TRUE, 0, sizeof(int), n_devices_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], gpu_id_gpu[index_devices], CL_TRUE, 0, sizeof(int), gpu_id_pointer[index_devices], 0, NULL, NULL);
    }

#elif (PRECISION == PREC_DOUBLE)

    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      clEnqueueWriteBuffer(command_queue[index_devices], X_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n * *d, X_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_gpu[index_devices], CL_TRUE, 0, sizeof(int), n, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], d_gpu[index_devices], CL_TRUE, 0, sizeof(int), d, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, kpar_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], exp_kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, exp_kpar_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], OMEGA_gpu[index_devices], CL_TRUE, 0, sizeof(double), OMEGA_double, 0, NULL, NULL);
      //      clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n, v_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], split_row_matrix_gpu[index_devices], CL_TRUE, 0, sizeof(int), split_row_matrix_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], n_devices_gpu[index_devices], CL_TRUE, 0, sizeof(int), n_devices_pointer[index_devices], 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], gpu_id_gpu[index_devices], CL_TRUE, 0, sizeof(int), gpu_id_pointer[index_devices], 0, NULL, NULL);
    }

#endif
  

  // *******************************  
  for(i = 0; i < *npar; i++) exp_kpar[i] = exp(kpar[i]);

  // Initialize random number generator
  const gsl_rng_type *type_random_sampling;
  gsl_rng *uniform_random_variable;

  type_random_sampling = gsl_rng_mt19937;
  uniform_random_variable = gsl_rng_alloc (type_random_sampling);
  gsl_rng_set(uniform_random_variable, 2149 + *SEED * 13);
  // **********************************

  // Initialize Bernoulli random vector r
  double tmp, tmp2;
  /* double *rvect = (double*) malloc (*n * sizeof(double)); */
  // Draw a random vector r
  for(i = 0; i < *n; i++) {
    for(j = 0; j < *NRVECT; j++) {
      tmp = gsl_rng_uniform(uniform_random_variable);
      if(tmp > 0.5) rvect[i + j * (*n)] = 1;
      else rvect[i + j * (*n)] = -1;
    }
  }
  // **********************

  double *gradient = (double*) malloc (*n * sizeof(double));

  double *K_inv_y = (double*) malloc (*n * 2 * sizeof(double));
  double *K_inv_r = (double*) malloc (*n * (*NRVECT) * sizeof(double));
  double *init_solution_K_inv_y = (double*) malloc (*n * sizeof(double));
  double *init_solution_K_inv_r = (double*) malloc (*n * (*NRVECT) * sizeof(double));
  for(i = 0; i < *n; i++) {
    for(j = 0; j < 2; j++) {
      K_inv_y[i + j * *n] = 0;
    }
	init_solution_K_inv_y[i] = 0;
  }

  for(i = 0; i < *n; i++) {
    for(j = 0; j < *NRVECT; j++) {
      K_inv_r[i + j * *n] = 0;
	  init_solution_K_inv_r[i + j * *n] = 0;
    }
  }

  compute_gradient(X, y, n, d, kpar, exp_kpar, npar, OMEGA, uniform_random_variable, priors, rvect, K_inv_y, K_inv_r, NRVECT, gradient, *THRS_CONVERGENCE_CG_pointer, *THRS_EARLY_STOP_CG_pointer, *DECAY_EARLY_STOP_CG_pointer, command_queue, program, kernel_kernel_matrix_vector_multiplication, kernel_dK_vector_multiplication_psi_sigma, kernel_dK_vector_multiplication_psi_tau, v_gpu, res_gpu, res_double, init_solution_K_inv_r, init_solution_K_inv_y);

  for(i = 0; i < *npar; i++) {
    SAMPLES_GRAD[i * (*MAXIT + 1)] = gradient[i];
    SAMPLES[i * (*MAXIT + 1)] = kpar[i];
  }

  // Initialize stochastic gradient schedule parameters a and b - epsilon = a (b + t)^(-gamma) where t is the iteration number. Parameters a and b are set based on (i) min and max values of epsilon, (ii) max iteration number, and (iii) gamma
  double tmp_schedule = pow(*EPSILON_MAX / *EPSILON_MIN, - 1 / *GAMMA);
  double b_schedule = (*MAXIT * tmp_schedule - 1) / (1 - tmp_schedule);
  double a_schedule = *EPSILON_MAX * pow(b_schedule + 1, *GAMMA);
  // *******************************
  
  char filename[80];

  snprintf(filename, sizeof(filename), "%s_SEED_%d.txt", OUTPUT_FILE, *SEED);
  FILE* file = fopen(filename, "w");

  snprintf(filename, sizeof(filename), "%s_SEED_%d.txt", OUTPUT_FILE_GRAD, *SEED);
  FILE* file_GRAD = fopen(filename, "w");

  snprintf(filename, sizeof(filename), "%s_SEED_%d.txt", OUTPUT_FILE_RATIO_VAR, *SEED);
  FILE* file_RATIO_VAR = fopen(filename, "w");

  snprintf(filename, sizeof(filename), "%s_SEED_%d.txt", OUTPUT_FILE_EPSILON, *SEED);
  FILE* file_EPSILON = fopen(filename, "w");

  // Stochastic Gradient Loop
  int freeze_flag = 0;
  int freeze_cov_flag = 0;
  gsl_matrix* V_gradients = gsl_matrix_alloc(*npar, *npar);
  gsl_matrix* M12_V_M12 = gsl_matrix_alloc(*npar, *npar);
  double epsilon;
  for(int t = 0; t < *MAXIT; t++) {
    
    if((t % 1) == 0) { printf("\n%d    ", t); fflush(stdout); }
    
	// Check whether the variance of the injected noise is larger than the variance of the noisy gradients 
	if(freeze_flag == 0) {

	  epsilon = a_schedule * pow(b_schedule + (t + 1), -*GAMMA);
	  
	  // Calculate the covariance of the noisy gradients
	  if(t == 0) {
		for(i = 0; i < *npar; i++) {
		  for(j = 0; j < *npar; j++) {
			if(i == j) gsl_matrix_set(V_gradients, i, j, 1);
			else gsl_matrix_set(V_gradients, i, j, 0);
		  }
		}
	  }

	  else {
		
		int min_ind_to_estimate = 0;
		if(t > 100) min_ind_to_estimate = t - 100;

		double mean_grad[*npar];
		for(i = 0; i < *npar; i++) {
		  mean_grad[i] = 0;
		  for(int k = min_ind_to_estimate; k <= t; k++) {
			mean_grad[i] += SAMPLES_GRAD[k + i * (*MAXIT + 1)];
		  }
		  mean_grad[i] /= (double) (t - min_ind_to_estimate + 1);
		}

		double tmp_cum;
		for(i = 0; i < *npar; i++) {
		  for(j = 0; j < *npar; j++) {
			tmp_cum = 0;
			for(int k = min_ind_to_estimate; k <= t; k++) {
			  tmp_cum += (SAMPLES_GRAD[k + i * (*MAXIT + 1)] - mean_grad[i]) * (SAMPLES_GRAD[k + j * (*MAXIT + 1)] - mean_grad[j]);
			}
			tmp_cum /= (double) (t - min_ind_to_estimate + 1);
			gsl_matrix_set(V_gradients, i, j, tmp_cum);
		  }
		}
	  }

	  double tmp_cum;
	  for(i = 0; i < *npar; i++) {
		for(j = 0; j < *npar; j++) {
		  tmp_cum = 0;
		  for(int k = 0; k < *npar; k++) {
			tmp_cum += M12[i + k * *npar] * gsl_matrix_get(V_gradients, k, j);
		  }
		  gsl_matrix_set(M12_V_M12, i, j, tmp_cum);
		}
	  }

	  for(i = 0; i < *npar; i++) {
		for(j = 0; j < *npar; j++) {
		  tmp_cum = 0;
		  for(int k = 0; k < *npar; k++) {
			tmp_cum += gsl_matrix_get(M12_V_M12, i, k) * M12[k + j * *npar];
		  }
		  gsl_matrix_set(M12_V_M12, i, j, tmp_cum);
		}
	  }

	  gsl_eigen_symmv_workspace *eigensystem = gsl_eigen_symmv_alloc(*npar);
	  gsl_vector *eval = gsl_vector_alloc (*npar);
	  gsl_matrix *evec = gsl_matrix_alloc (*npar, *npar);
	  gsl_eigen_symmv (M12_V_M12, eval, evec, eigensystem);
	  gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_DESC);

	  double eigen_M12_V_M12 = gsl_vector_get(eval, 0);
	  double ratio_var = epsilon / 4 * eigen_M12_V_M12;

	  printf("XX ratio vars %lf XX", ratio_var);
	  fprintf(file_RATIO_VAR, "%lf\n", ratio_var);

	  if(t > 100) {
		if(ratio_var < 0.01) {
		  freeze_flag = 1;
		  printf("\n\n XX FREEZE XX %d XX FREEZE XX\n\n", t);
		}
	  }
	  
	  /* for(i = 0; i < *npar; i++) { */
	  /* 	for(j = 0; j < *npar; j++) { */
	  /* 	  printf("XX %lf XX", gsl_matrix_get(M12_V_M12, i, j)); */
	  /* 	} */
	  /* } */

	}

	// Draw some random components of random vector r - with probability P = 1% redraw r[i]
	if((t % 20) == 0) {
	  for(i = 0; i < *n; i++) {
		for(j = 0; j < *NRVECT; j++) {
		  tmp2 = gsl_rng_uniform(uniform_random_variable);
		  if(tmp2 < 0.5) {
			tmp = gsl_rng_uniform(uniform_random_variable);
			if(tmp > 0.5) rvect[i + j * *n] = 1;
			else rvect[i + j * *n] = -1;
		  }
		}
	  } 
	}
	// **********************
	
	compute_gradient(X, y, n, d, kpar, exp_kpar, npar, OMEGA, uniform_random_variable, priors, rvect, K_inv_y, K_inv_r, NRVECT, gradient, *THRS_CONVERGENCE_CG_pointer, *THRS_EARLY_STOP_CG_pointer, *DECAY_EARLY_STOP_CG_pointer, command_queue, program, kernel_kernel_matrix_vector_multiplication, kernel_dK_vector_multiplication_psi_sigma, kernel_dK_vector_multiplication_psi_tau, v_gpu, res_gpu, res_double, init_solution_K_inv_r, init_solution_K_inv_y);
	
	printf("   %lf %lf %lf     ", kpar[0], kpar[1], kpar[2]);
	/* printf("%lf %lf %lf    ", diag_M[0], diag_M[1], diag_M[2]); */
	printf("%lf %lf %lf\n", gradient[0], gradient[1], gradient[2]);
	
	double M12_grad[*npar];
	for(i = 0; i < *npar; i++) {
	  M12_grad[i] = 0;
	  for(j = 0; j < *npar; j++) {
		M12_grad[i] += M12[i + j * *npar] * gradient[j];
	  }
	}

	double M_grad[*npar];
	for(i = 0; i < *npar; i++) {
	  M_grad[i] = 0;
	  for(j = 0; j < *npar; j++) {
		M_grad[i] += M12[i + j * *npar] * M12_grad[j];
	  }
	}

	double eta[*npar];
	double zeta[*npar];
	for(i = 0; i < *npar; i++) zeta[i] = gsl_ran_gaussian(uniform_random_variable, 1.0);

	for(i = 0; i < *npar; i++) {
	  eta[i] = 0;
	  for(j = 0; j < *npar; j++) {
		eta[i] += M12[i + j * *npar] * zeta[j];
	  }
	  eta[i] *= sqrt(epsilon);
	}
	

	norm_kpar_flag = 1;
	for(i = 0; i < *npar; i++) {
	  kpar_tmp[i] = kpar[i] + epsilon / 2 * M_grad[i] + eta[i];

	  norm_kpar_flag *= (abs(kpar_tmp[i]) < 10);
	}
	
	//	if(norm_kpar_flag == 1) {
	for(i = 0; i < *npar; i++) { 
	  kpar[i] = kpar_tmp[i];
	  exp_kpar[i] = exp(kpar[i]);
	}
	//	}

	for(i = 0; i < *npar; i++) { 
	  fprintf(file, "%lf ", kpar[i]);
	  fprintf(file_GRAD, "%lf ", gradient[i]);
	  /* fprintf(file_DIAG_G, "%lf ", diag_G[i]); */
	  
	  SAMPLES[t + 1 + i * (*MAXIT + 1)] = kpar[i];
	  SAMPLES_GRAD[t + 1 + i * (*MAXIT + 1)] = gradient[i];
	  /* SAMPLES_DIAG_G[t + 1 + i * (*MAXIT + 1)] = diag_G[i]; */
	}

	fprintf(file, "\n");
	fprintf(file_GRAD, "\n");
	/* fprintf(file_DIAG_G, "\n"); */
	fprintf(file_EPSILON, "%.10lf\n", epsilon);

	for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
      clEnqueueWriteBuffer(command_queue[index_devices], kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, kpar_double, 0, NULL, NULL);
      clEnqueueWriteBuffer(command_queue[index_devices], exp_kpar_gpu[index_devices], CL_TRUE, 0, sizeof(double) * 3, exp_kpar_double, 0, NULL, NULL);
	}
    
  }
  // ************************

  fclose(file);
  fclose(file_GRAD);
  /* fclose(file_DIAG_G); */
  fclose(file_EPSILON);
  
  printf("\n%d    ", *MAXIT);
  compute_gradient(X, y, n, d, kpar, exp_kpar, npar, OMEGA, uniform_random_variable, priors, rvect, K_inv_y, K_inv_r, NRVECT, gradient, *THRS_CONVERGENCE_CG_pointer, *THRS_EARLY_STOP_CG_pointer, *DECAY_EARLY_STOP_CG_pointer, command_queue, program, kernel_kernel_matrix_vector_multiplication, kernel_dK_vector_multiplication_psi_sigma, kernel_dK_vector_multiplication_psi_tau, v_gpu, res_gpu, res_double, init_solution_K_inv_r, init_solution_K_inv_y);
  for(i = 0; i < *npar; i++) 
    gradient_final[i] = gradient[i];


  /* free(rvect); */
  free(gradient);
  free(K_inv_y);
  free(K_inv_r);
  free(init_solution_K_inv_y);
  free(init_solution_K_inv_r);

  gsl_matrix_free(V_gradients);

#if (PRECISION == PREC_FLOAT)

  free(X_float);
  free(kpar_float);
  free(exp_kpar_float);
  free(OMEGA_float);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    free(res_float[index_devices]);
  }

  free(x_float);
  free(p_float);
  free(Ax_float);
  free(Ap_float);

#elif (PRECISION == PREC_DOUBLE)

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    free(res_double[index_devices]);
  }

#endif

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
    clReleaseMemObject(X_gpu[index_devices]);
 
    clReleaseMemObject(d_gpu[index_devices]);
    clReleaseMemObject(n_gpu[index_devices]);
    clReleaseMemObject(kpar_gpu[index_devices]);
    clReleaseMemObject(exp_kpar_gpu[index_devices]);
    clReleaseMemObject(OMEGA_gpu[index_devices]);
    clReleaseMemObject(v_gpu[index_devices]);
    clReleaseMemObject(split_row_matrix_gpu[index_devices]);
    clReleaseMemObject(n_devices_gpu[index_devices]);
    clReleaseMemObject(gpu_id_gpu[index_devices]);
    clReleaseMemObject(res_gpu[index_devices]);

    clReleaseProgram(program[index_devices]);
    clReleaseKernel(kernel_kernel_matrix_vector_multiplication[index_devices]);
    clReleaseKernel(kernel_dK_vector_multiplication_psi_sigma[index_devices]);
    clReleaseKernel(kernel_dK_vector_multiplication_psi_tau[index_devices]);
    clReleaseCommandQueue(command_queue[index_devices]);
    clReleaseContext(context[index_devices]);
  }


  return(0);
}
