int dK_vector_multiplication(double *X, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double* v, int component, double *y, cl_command_queue *command_queue, cl_program *program, cl_kernel *kernel_dK_vector_multiplication_psi_sigma, cl_kernel *kernel_dK_vector_multiplication_psi_tau, cl_mem *v_gpu, cl_mem *res_gpu, double **res_double) {
  
  cl_int err;
  int i, j;

  double *v_double = v;

  // load data into the input buffer
#if (PRECISION == PREC_FLOAT)

    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n, v_float, 0, NULL, NULL);
    }

#elif (PRECISION == PREC_DOUBLE)

    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n, v_double, 0, NULL, NULL);
    }

#endif

  size_t global[2];
  global[0] = *n;
  global[1] = SPLIT_ROW_MATRIX;

  // enqueue the kernel command for execution
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++) {
    if(component == 0) clEnqueueNDRangeKernel(command_queue[index_devices], kernel_dK_vector_multiplication_psi_sigma[index_devices], 2, NULL, &global[0], NULL, 0, NULL, NULL);
    if(component > 1) clEnqueueNDRangeKernel(command_queue[index_devices], kernel_dK_vector_multiplication_psi_tau[index_devices], 2, NULL, &global[0], NULL, 0, NULL, NULL);
  }

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clFinish(command_queue[index_devices]);

#if (PRECISION == PREC_FLOAT)
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n * SPLIT_ROW_MATRIX, res_float[index_devices], 0, NULL, NULL);
#elif (PRECISION == PREC_DOUBLE)
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n * SPLIT_ROW_MATRIX, res_double[index_devices], 0, NULL, NULL);
#endif

#if (PRECISION == PREC_FLOAT)

  for(i = 0; i < *n; i++) {
    /* printf("%lf %lf \n",res_float[i], res_float[i + *n]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
	y[i] += res_float[index_devices][i + j * *n];
      }
    }

  }

#elif (PRECISION == PREC_DOUBLE)

  for(i = 0; i < *n; i++) {
    /* printf("%lf \n", res_double[0][i]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
	y[i] += res_double[index_devices][i + j * *n];
      }
    }
  }

#endif

  /* free(v_double); */
  
  return 0;
}
