void compute_gradient(double *X, double *y, int *n, int *d, double *kpar, double *exp_kpar, int *npar, double *OMEGA, gsl_rng *uniform_random_variable, double *priors, double *rvect, double *K_inv_y, double *K_inv_r, int *NRVECT, double *gradient, double THRS_CONVERGENCE_CG, double THRS_EARLY_STOP_CG, double DECAY_EARLY_STOP_CG, cl_command_queue *command_queue, cl_program *program, cl_kernel *kernel_kernel_vector_multiplication, cl_kernel *kernel_dK_vector_multiplication_psi_sigma, cl_kernel *kernel_dK_vector_multiplication_psi_tau, cl_mem *v_gpu, cl_mem *res_gpu, double **res_double, double *init_solution_K_inv_r, double *init_solution_K_inv_y) {

  int i, j;

  double grad_part1 = 0, grad_part2 = 0, grad_part3;

  // Compute an unbiased version of K^-1 r and two unbiased versions of K^-1 y
  for(i = 0; i < *NRVECT; i++) {
    solve_CG(X, n, d, kpar, exp_kpar, OMEGA, &rvect[0 + i * *n], &init_solution_K_inv_r[0 + i * *n], uniform_random_variable, 1, &K_inv_r[0 + i * *n], THRS_CONVERGENCE_CG, THRS_EARLY_STOP_CG, DECAY_EARLY_STOP_CG, command_queue, program, kernel_kernel_vector_multiplication, v_gpu, res_gpu, res_double);
  }

  solve_CG(X, n, d, kpar, exp_kpar, OMEGA, y, init_solution_K_inv_y, uniform_random_variable, 2, K_inv_y, THRS_CONVERGENCE_CG, THRS_EARLY_STOP_CG, DECAY_EARLY_STOP_CG, command_queue, program, kernel_kernel_vector_multiplication, v_gpu, res_gpu, res_double);
  double *K_inv_y_1 = &K_inv_y[0];
  double *K_inv_y_2 = &K_inv_y[*n];

  double *dK_r = (double*) malloc (*n * *NRVECT * sizeof(double));
  double *dK_K_inv_y = (double*) malloc (*n * sizeof(double));
  
  // Compute dK r and dK K^-1 y for all parameters and use them to compute the components of the gradient
  for(i = 0; i < *npar; i++) {
    // Derivatives wrt psi_sigma and psi_tau on the GPU
    if(i != 1) {
      for(j = 0; j < *NRVECT; j++) {
	dK_vector_multiplication(X, n, d, kpar, exp_kpar, OMEGA, &rvect[0 + j * *n], i, &dK_r[0 + j * *n], command_queue, program, kernel_dK_vector_multiplication_psi_sigma, kernel_dK_vector_multiplication_psi_tau, v_gpu, res_gpu, res_double);
      }

      dK_vector_multiplication(X, n, d, kpar, exp_kpar, OMEGA, K_inv_y_1, i, dK_K_inv_y, command_queue, program, kernel_dK_vector_multiplication_psi_sigma, kernel_dK_vector_multiplication_psi_tau, v_gpu, res_gpu, res_double);
    }
    
    // Derivative wrt psi_lambda
    else {
      for(j = 0; j < *n; j++) { 
	for(int k = 0; k < *NRVECT; k++) {
	  dK_r[j + k * *n] = exp_kpar[1] * rvect[j + k * *n];
	}
      }
      for(j = 0; j < *n; j++) { dK_K_inv_y[j] = exp_kpar[1] * K_inv_y[j]; }
    }
    
    grad_part1 = 0;
    for(j = 0; j < *n; j++) {
      for(int k = 0; k < *NRVECT; k++) {
	grad_part1 += dK_r[j + k * *n] * K_inv_r[j + k * *n];
      }
    }
    grad_part1 /= (double) (*NRVECT);
    grad_part1 *= -0.5;
    
    grad_part2 = 0;
    for(j = 0; j < *n; j++) grad_part2 += dK_K_inv_y[j] * K_inv_y_2[j];
    grad_part2 *= 0.5;
    
    grad_part3 = 0;
    grad_part3 = priors[0 + i * 2] - priors[1 + i * 2] * exp_kpar[i];
    gradient[i] = grad_part1 + grad_part2 + grad_part3;
  }

  free(dK_r);
  free(dK_K_inv_y);
}
