

#if (PRECISION == PREC_FLOAT)

void kernel_matrix_vector_multiplication(int *n, cl_command_queue *command_queue, cl_kernel *kernel, cl_mem *v_gpu, float *v_float, cl_mem *res_gpu, float **res_float, float *exp_kpar_float, float *OMEGA_float, float *y) {

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
	clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n, v_float, 0, NULL, NULL);

  size_t global[2];
  global[0] = *n;
  global[1] = SPLIT_ROW_MATRIX;

  // enqueue the kernel command for execution
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueNDRangeKernel(command_queue[index_devices], kernel[index_devices], 2, NULL, &global[0], NULL, 0, NULL, NULL);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clFinish(command_queue[index_devices]);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(float) * *n * SPLIT_ROW_MATRIX, res_float[index_devices], 0, NULL, NULL);

  int i, j;

  for(i = 0; i < *n; i++) {
    /* printf("%lf %lf \n",res_float[i], res_float[i + *n]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
		y[i] += res_float[index_devices][i + j * *n];
      }
    }
    y[i] += (exp_kpar_float[1] + *OMEGA_float) * v_float[i];
  }

}


#elif (PRECISION == PREC_DOUBLE)

void kernel_matrix_vector_multiplication(int *n, cl_command_queue *command_queue, cl_kernel *kernel, cl_mem *v_gpu, double *v_double, cl_mem *res_gpu, double **res_double, double *exp_kpar, double *OMEGA, double *y) {

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueWriteBuffer(command_queue[index_devices], v_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n, v_double, 0, NULL, NULL);

  size_t global[2];
  global[0] = *n;
  global[1] = SPLIT_ROW_MATRIX;

  // enqueue the kernel command for execution
  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueNDRangeKernel(command_queue[index_devices], kernel[index_devices], 2, NULL, &global[0], NULL, 0, NULL, NULL);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clFinish(command_queue[index_devices]);

  for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)
    clEnqueueReadBuffer(command_queue[index_devices], res_gpu[index_devices], CL_TRUE, 0, sizeof(double) * *n * SPLIT_ROW_MATRIX, res_double[index_devices], 0, NULL, NULL);

  int i, j;

  for(i = 0; i < *n; i++) {
    /* printf("%lf %lf \n",res_double[i], res_double[i + *n]); */
    y[i] = 0;
    for(int index_devices = 0; index_devices < N_DEVICES; index_devices++)  {
      for(j = 0; j < SPLIT_ROW_MATRIX; j++) {
		y[i] += res_double[index_devices][i + j * *n];
      }
    }
    y[i] += (exp_kpar[1] + *OMEGA) * v_double[i];
  }

}

#endif



int solve_CG(double *X, int *n, int *d, double *kpar, double *exp_kpar, double *OMEGA, double *b_double, double *init_solution, gsl_rng *uniform_random_variable, int n_results, double *result, double THRS_CONVERGENCE_CG, double THRS_EARLY_STOP_CG, double DECAY_EARLY_STOP_CG, cl_command_queue *command_queue, cl_program *program, cl_kernel *kernel_kernel_matrix_vector_multiplication, cl_mem *v_gpu, cl_mem *res_gpu, double **res_double) {

  gsl_vector* b = gsl_vector_alloc(*n);
  for(int i = 0; i < *n; i++) gsl_vector_set(b, i, b_double[i]);
  
  double u;
  
  int j;
  
  int n_after_early_stop[n_results];
  for(int i = 0; i < n_results; i++) {
    for(j = 0; j < MAX_LOOPS_CG; j++) {
      u = gsl_rng_uniform(uniform_random_variable);	  
      if(u > exp(-j / DECAY_EARLY_STOP_CG)) break;
    }
    n_after_early_stop[i] = j;

    /* printf("N EARLY STOP %d \n", n_after_early_stop[i]); */
  }

  double max_weighting_factor = 0, tmp_weighting_factor = 0;
  for(int i = 0; i < n_results; i++) {
	tmp_weighting_factor = exp( (n_after_early_stop[i] * (n_after_early_stop[i]-1)) / 2 / DECAY_EARLY_STOP_CG); // sum_{i=0}^(n-1) i = n * (n-1)/2
    if(tmp_weighting_factor > max_weighting_factor) max_weighting_factor = tmp_weighting_factor;
  }

  THRS_EARLY_STOP_CG = sqrt(*n) / max_weighting_factor;
    
  int earlystop = 0, index_expansion = 0;
  
  // Initialize x using the vector in result (that is what is passed to the function as the value of K^-1 b from the previous iteration)
  gsl_vector* x = gsl_vector_alloc(*n);
  gsl_vector* result_gsl = gsl_vector_alloc(*n);
  for(int i = 0; i < *n; i++) gsl_vector_set(result_gsl, i, init_solution[i]);
  gsl_vector_memcpy(x, result_gsl);
  
  
#if (PRECISION == PREC_FLOAT)
  
  float *x_float;
  x_float = (float*) malloc (*n * sizeof(float));
  for(int i = 0; i < *n; i++) x_float[i] = (float) gsl_vector_get(x, i);
  
#elif (PRECISION == PREC_DOUBLE)
  
  double *x_double;
  x_double = (double*) malloc (*n * sizeof(double));
  for(int i = 0; i < *n; i++) x_double[i] = gsl_vector_get(x, i);
  
#endif
  
  gsl_vector* x_unbiased = gsl_vector_alloc(*n);
  gsl_vector* p_tmp = gsl_vector_alloc(*n);
  double tmp;
  
  gsl_vector* Ax = gsl_vector_alloc(*n);
  
#if (PRECISION == PREC_FLOAT)
  float *Ax_float;
  Ax_float = (float*) malloc (*n * sizeof(float));
  
  kernel_matrix_vector_multiplication(n, command_queue, kernel_kernel_matrix_vector_multiplication, v_gpu, x_float, res_gpu, res_float, exp_kpar_float, OMEGA_float, Ax_float);
  for(int i = 0; i < *n; i++) gsl_vector_set(Ax, i, (double) Ax_float[i]);
  
#elif (PRECISION == PREC_DOUBLE)
  
  double *Ax_double;
  Ax_double = (double*) malloc (*n * sizeof(double));
  
  kernel_matrix_vector_multiplication(n, command_queue, kernel_kernel_matrix_vector_multiplication, v_gpu, x_double, res_gpu, res_double, exp_kpar, OMEGA, Ax_double);
  for(int i = 0; i < *n; i++) gsl_vector_set(Ax, i, Ax_double[i]);
  
#endif


  /* // r = b - Ax */
  gsl_vector* r = gsl_vector_alloc(*n);
  gsl_vector_memcpy(r, b);
  gsl_vector_sub(r, Ax);

  gsl_vector* q = gsl_vector_alloc(*n);

  // p = b - Ax
  gsl_vector* p = gsl_vector_alloc(*n);
  gsl_vector_memcpy(p, r);

#if (PRECISION == PREC_FLOAT)
  float *p_float;
  p_float = (float*) malloc (*n * sizeof(float));
  float *Ap_float;
  Ap_float = (float*) malloc (*n * sizeof(float));
#elif (PRECISION == PREC_DOUBLE)
  double *p_double;
  p_double = (double*) malloc (*n * sizeof(double));
  double *Ap_double;
  Ap_double = (double*) malloc (*n * sizeof(double));
#endif

  double diff;
  double alphaN = 0;
  double alphaD = 0;
  double alpha = 0;
  double betaN = 0;
  double beta = 0;
  gsl_vector* Ap = gsl_vector_alloc(*n);
  
  int k = 0;
  do {
    // alpha = rTr / pAp
    gsl_blas_ddot(r, r, &alphaN);
	
#if (PRECISION == PREC_FLOAT)
    for(int i = 0; i < *n; i++) p_float[i] = (float) gsl_vector_get(p, i);
    kernel_matrix_vector_multiplication(n, command_queue, kernel_kernel_matrix_vector_multiplication, v_gpu, (float *) p_float, res_gpu, res_float, exp_kpar_float, OMEGA_float, Ap_float);
    for(int i = 0; i < *n; i++) gsl_vector_set(Ap, i, (double) Ap_float[i]);
#elif (PRECISION == PREC_DOUBLE)
    for(int i = 0; i < *n; i++) p_double[i] = gsl_vector_get(p, i);
    kernel_matrix_vector_multiplication(n, command_queue, kernel_kernel_matrix_vector_multiplication, v_gpu, p_double, res_gpu, res_double, exp_kpar, OMEGA, Ap_double);
    for(int i = 0; i < *n; i++) gsl_vector_set(Ap, i, Ap_double[i]);
#endif
    
    gsl_blas_ddot(p, Ap, &alphaD);
    alpha = alphaN / alphaD;
    
    gsl_vector_memcpy(q, r);
    
    // x = x + alpha*p
    gsl_vector_scale(p, alpha);
    gsl_vector_add(x, p);
    
    // r = r - alpha*Ap
    gsl_vector_scale(Ap, alpha);
    gsl_vector_sub(r, Ap);
    
    gsl_vector_sub(q, r);
    gsl_blas_ddot(q, q, &diff);
    
    if(sqrt(diff) < THRS_CONVERGENCE_CG) {
      /* gsl_vector_memcpy(result_gsl, x); */
      for(unsigned int i = 0; i < n_results; i++) {
		for(unsigned int j = 0; j < *n; j++)
		  result[j + i * *n] = gsl_vector_get(x, j);
      }
      break;
    }
    
    if(earlystop == 0) {
      if(sqrt(diff) < THRS_EARLY_STOP_CG) {
		earlystop = 1;
		gsl_vector_memcpy(x_unbiased, x);
		gsl_vector_sub(x_unbiased, p);
		index_expansion = 0;
      }
    }
    
    if(earlystop == 1) {
      int checkpoint = 1;
      for(unsigned int i = 0; i < n_results; i++) checkpoint *= (index_expansion >= n_after_early_stop[i]);
      if(checkpoint == 1) {
		break;
      }
      
      else {
		gsl_vector_memcpy(p_tmp, p);
		/* tmp = 1.0; */ // Explicit (old fashioned) way to compute the weighting factor for the terms in the expansion
		/* for(unsigned int i = 0; i < index_expansion; i++) tmp *= exp(i / DECAY_EARLY_STOP_CG); */
		
		/* printf("%d %lf %lf\n", index_expansion, tmp, exp(index_expansion * (index_expansion-1) / 2 / DECAY_EARLY_STOP_CG) ); */

		tmp = exp(index_expansion * (index_expansion-1) / 2 / DECAY_EARLY_STOP_CG); // Weighting factor for the terms in the expansion - prod_i exp(i / DECAY_EARLY_STOP_CG) = exp((sum_i i) / DECAY_EARLY_STOP_CG)

		gsl_vector_scale(p_tmp, tmp);
		gsl_vector_add(x_unbiased, p_tmp);
		/* gsl_vector_memcpy(result_gsl, x_unbiased); */
		
		for(unsigned int i = 0; i < n_results; i++) {
		  if( index_expansion == (n_after_early_stop[i]-1) ) {
			for(unsigned int j = 0; j < *n; j++) {
			  result[j + i * *n] = gsl_vector_get(x_unbiased, j);
			}
		  }
		}
      }
      
      index_expansion++;
    }
    
    // beta = (r1Tr1 / r0Tr0) * alpha^(-1)
    gsl_blas_ddot(r, r, &betaN);
    beta = (betaN * alphaD) / (alphaN * alphaN);
    
    // p = beta*p + r
    gsl_vector_scale(p, beta);
    gsl_vector_add(p, r);
    k++;
    
  } while(k < MAX_LOOPS_CG);
  
  if(k == MAX_LOOPS_CG) {
    gsl_vector_memcpy(result_gsl, x);
    printf("CG reached limit of %d \n", MAX_LOOPS_CG);
    for(int i = 0; i < *n; i++) result[i] = gsl_vector_get(result_gsl, i);
  }
  
  for(int i = 0; i < *n; i++) init_solution[i] = gsl_vector_get(x, i);
  
  /* printf("\n\nFinished in %d iterations\n\n", k); */
  printf("%d ", k + 2);


  free(x_double);
  free(p_double);
  free(Ax_double);
  free(Ap_double);
  
  gsl_vector_free(r);
  gsl_vector_free(x);
  gsl_vector_free(x_unbiased);
  gsl_vector_free(p);
  gsl_vector_free(p_tmp);
  gsl_vector_free(Ap);

  gsl_vector_free(Ax);
  gsl_vector_free(result_gsl);

  return 0;
}
